// This file is part of noxos and licensed under the MIT open source license

#include "utils/logger.h"
#include "utils/core.h"
#include "utils/memory.h"
#include "boot/boot_info.h"
#include "boot/config.h"
#include "platform/interrupts.h"
#include "platform/gdt.h"
#include "platform/syscall.h"
#include "mm/page_frame.h"
#include "mm/page_map.h"
#include "mm/region.h"
#include "drivers/builtin/time/pit.h"
#include "drivers/builtin/graphics/renderer.h"
#include "drivers/builtin/fs/vfs.h"
#include "drivers/builtin/acpi/acpi.h"
#include "drivers/builtin/drive_manager.h"
#include "drivers/builtin/pci.h"
#include "drivers/builtin/ps2/controller.h"
#include "drivers/builtin/tty.h"
#include "drivers/driver.h"
#include "proc/scheduler.h"

void limine_terminal_print(boot_info_T* boot_info, string_t string) {
    boot_info->terminal->write(boot_info->terminal->terminals[0], string, string_length(string));
}

void kernel_init(boot_info_T* boot_info) {
    pframe_manager_init(boot_info);

    CORE_INTERRUPTS_DISABLE
    gdt_init();
    idt_init();
    pic_init();
    pit_set_divisor(PIT_DIVISOR);

    paging_init();
    CORE_INTERRUPTS_ENABLE

    memory_allocator_init((void*)MEM_REGION_KERNEL_HEAP);

    graphics_renderer_init(boot_info);

    vfs_init(boot_info);

    sysconfig_init();

    if (g_sysconfig->log_graphical) graphical_log_init();

    scheduler_init(boot_info);

    driver_manager_init();

    acpi_init(boot_info);

    drive_manager_init();

    pci_init();

    tty_init();

    ps2_controller_init();
}

void kmain(boot_info_T boot_info) {

    limine_terminal_print(&boot_info, "Booting NoxOS...\n");
    log(LOG_DEBUG, "Booting NoxOS");

    kernel_init(&boot_info);

    log(LOG_INFO, "!=====[ Kernel Initialized ]=====!\n");

    file_descriptor_t fd;

    syscall_perform(SYSCALL_NX_FS_OPEN, (uint64_t)"/initrd/test.elf", 0, &fd, 0);

    pid_t pid;
    process_config_T conf = {
            .privilege_level = 0,
            .executable      = fd,
            .num_inherit_fds = 0
    };
    memory_copy("test", conf.name, 5);

    syscall_perform(SYSCALL_NX_PROC_CREATE, (uint64_t)&conf, (uint64_t)&pid, 0, 0);

    process_T* process = scheduler_get_process(pid);

    scheduler_dump_info(scheduler_get_process(PROCESS_KERNEL), 0);

    g_tty->output   = &process->stdin;
    process->stdout = &g_tty->input;

    syscall_perform(SYSCALL_NX_PROC_SIGNAL_SEND, pid, PROCESS_SIGNAL_START, 0, 0);


    CORE_HALT_FOREVER
}
