// This file is part of noxos and licensed under the MIT open source license

#include "drivers/builtin/acpi/acpi.h"
#include "drivers/builtin/acpi/rsdp.h"
#include "drivers/builtin/acpi/fadt.h"
#include "drivers/builtin/acpi/mcfg.h"
#include "utils/logger.h"
#include "utils/memory.h"
#include "utils/panic.h"

acpi_fadt_T*  g_acpi_table_fadt;
acpi_mcfg_T*  g_acpi_table_mcfg;

void acpi_init(boot_info_T* boot_info) {
    log(LOG_INFO, "Initializing ACPI  -  RSDP: 0x%x", boot_info->rsdp);
    rsdp_descriptor_v1_T* rsdp = boot_info->rsdp;

    switch (rsdp->revision) {
        case 0: {
            log(LOG_INFO, "    Version: 1.0");
            if (!rsdp_descriptor_v1_verify(boot_info->rsdp)) {
                log(LOG_WARNING, "<ACPI> RSDP header verification failed");
            }
            break;
        }
        case 2: {
            log(LOG_INFO, "    Version: 2.0+");
            if (!rsdp_descriptor_v2_verify(boot_info->rsdp)) {
                log(LOG_WARNING, "<ACPI> RSDP header verification failed");
            }

            acpi_sdt_header_T* xsdt = ((rsdp_descriptor_v2_T*)boot_info->rsdp)->xsdt_address;

            g_acpi_table_fadt = (acpi_fadt_T*)acpi_find_table(xsdt, "FACP");
            if (g_acpi_table_fadt == NULL) {
                panic(NULL, "<ACPI> FADT table not found");
            }
            log(LOG_INFO, "<ACPI> found FADT at address 0x%x", g_acpi_table_fadt);

            g_acpi_table_mcfg = (acpi_mcfg_T*)acpi_find_table(xsdt, "MCFG");
            if (g_acpi_table_mcfg == NULL) {
                panic(NULL, "<ACPI> MCFG table not found");
            }
            log(LOG_INFO, "<ACPI> found MCFG at address 0x%x", g_acpi_table_mcfg);

            break;
        }
        default: {
            log(LOG_ERROR, "<ACPI> failed to parse RSDP table (invalid revision)");
            break;
        }
    }
}

acpi_sdt_header_T* acpi_find_table(acpi_sdt_header_T* xsdt, string_t table_id) {
    uint32_t num_entries = (xsdt->length - sizeof(acpi_sdt_header_T)) / 8;

    uint64_t* sdt_list = (uint64_t*)&xsdt[1];

    for (int i = 0; i < num_entries; i++) {
        acpi_sdt_header_T* sdt = (acpi_sdt_header_T*)(sdt_list[i]);
        if (!memory_compare(sdt->signature, (void*)table_id, 4)) continue;
        return sdt;
    }
    log(LOG_WARNING, "<ACPI> table '%.4s' not found (no matching xsdt entry)", table_id);
    return NULL;
}