// This file is part of noxos and licensed under the MIT open source license

#include "drivers/builtin/drive_manager.h"
#include "drivers/builtin/ahci.h"
#include "utils/memory.h"
#include "utils/logger.h"

drive_manager_T g_drive_manager;

drive_chunk_T* drive_chunk_alloc(drive_chunk_T* prev) {
    drive_chunk_T* chunk = memory_allocate(sizeof(drive_chunk_T));

    chunk->drives_bitmap = bitmap_init(DRIVE_CHUNK_SIZE);
    chunk->num_drives    = 0;
    chunk->next          = NULL;
    chunk->prev          = prev;
    if (prev != NULL) {
        prev->next = chunk;
    }

    return chunk;
}


drive_partition_chunk_T* drive_partition_chunk_alloc(drive_partition_chunk_T* prev) {
    drive_partition_chunk_T* chunk = memory_allocate(sizeof(drive_partition_chunk_T));

    chunk->partitions_bitmap = bitmap_init(DRIVE_PARTITION_CHUNK_SIZE);
    chunk->num_partitions    = 0;
    chunk->next              = NULL;
    chunk->prev              = prev;
    if (prev != NULL) {
        prev->next = chunk;
    }

    return chunk;
}

void drive_partition_chunk_destruct(drive_partition_chunk_T* chunk) {
    bitmap_destruct(&chunk->partitions_bitmap);
    memory_free(chunk);
}

void drive_manager_init(void) {
    g_drive_manager.chunks = drive_chunk_alloc(NULL);
}

drive_T* drive_manager_add_drive(drive_controller_types_E controller_type, void* controller, void* device) {
    uint32_t num_chunks = 0;

    drive_chunk_T* chunk = g_drive_manager.chunks;
    while (chunk->num_drives == DRIVE_CHUNK_SIZE) {
        if (chunk->next == NULL) {
            drive_chunk_alloc(chunk);
        }
        chunk = chunk->next;
        num_chunks++;
    }

    for (int i = 0; i < DRIVE_CHUNK_SIZE; i++) {
        if (bitmap_get(&chunk->drives_bitmap, i)) continue;
        bitmap_set(&chunk->drives_bitmap, i, true);

        drive_T* drive = &chunk->drives[i];

        drive->id_in_chunk       = i;
        drive->id                = (num_chunks * DRIVE_CHUNK_SIZE) + i;
        drive->partitions_chunks = drive_partition_chunk_alloc(NULL);
        drive->controller_type   = controller_type;
        drive->controller        = controller;
        drive->device            = device;
        drive->chunk             = chunk;

        chunk->num_drives += 1;

        log(LOG_INFO, "Registered drive %d", drive->id);

        return drive;
    }

    // this should never happen
    return NULL;
}

void drive_manager_remove_drive(drive_T* drive) {
    if (drive == NULL) return;
    drive_chunk_T* chunk = drive->chunk;
    if (!bitmap_get(&chunk->drives_bitmap, drive->id_in_chunk)) return;

    drive_partition_chunk_T* partition_chunk = drive->partitions_chunks;
    while (partition_chunk != NULL) {
        drive_partition_chunk_T* next = partition_chunk->next;
        drive_partition_chunk_destruct(partition_chunk);
        partition_chunk = next;
    }

    chunk->num_drives -= 1;
    bitmap_set(&chunk->drives_bitmap, drive->id_in_chunk, false);
}

drive_partition_T* drive_manager_add_partition(drive_T* drive, uint32_t start_sector, uint32_t end_sector) {
    uint32_t num_chunks = 0;

    drive_partition_chunk_T* chunk = drive->partitions_chunks;
    while (chunk->num_partitions == DRIVE_PARTITION_CHUNK_SIZE) {
        if (chunk->next == NULL) {
            drive_partition_chunk_alloc(chunk);
        }
        chunk = chunk->next;
        num_chunks++;
    }

    for (int i = 0; i < DRIVE_PARTITION_CHUNK_SIZE; i++) {
        if (bitmap_get(&chunk->partitions_bitmap, i)) continue;
        bitmap_set(&chunk->partitions_bitmap, i, true);

        drive_partition_T* partition = &chunk->partitions[i];

        partition->id_in_chunk  = i;
        partition->id           = (num_chunks * DRIVE_CHUNK_SIZE) + i;
        partition->drive        = drive;
        partition->chunk        = chunk;
        partition->start_sector = start_sector;
        partition->end_sector   = end_sector;

        // TODO: set partition filesystem

        chunk->num_partitions += 1;

        log(LOG_INFO, "Registered partition %d in drive %d", partition->id, drive->id);

        return partition;
    }

    // this should never happen
    return NULL;
}

void drive_manager_remove_partition(drive_partition_T* partition) {
    if (partition == NULL) return;
    drive_partition_chunk_T* chunk = partition->chunk;
    if (!bitmap_get(&chunk->partitions_bitmap, partition->id_in_chunk)) return;

    chunk->num_partitions -= 1;
    bitmap_set(&chunk->partitions_bitmap, partition->id_in_chunk, false);
}
