// This file is part of noxos and licensed under the MIT open source license

#include "drivers/builtin/elf/elf.h"
#include "drivers/builtin/elf/symbol.h"
#include "drivers/builtin/elf/segment.h"
#include "utils/memory.h"
#include "utils/logger.h"
#include "utils/math.h"

bool elf_executable_validate(elf_executable_T* executable) {
    if (executable->header.identity[0] != 0x7F ||
        executable->header.identity[1] != 'E' ||
        executable->header.identity[2] != 'L' ||
        executable->header.identity[3] != 'F')
    {
        log(LOG_ERROR, "ELF <0x%x> is invalid (magic number mismatch)", executable);
        return false;
    }

    if (executable->header.identity[5] != ELF_ENDIAN_LITTLE) {
        log(LOG_ERROR, "ELF <0x%x> is invalid (little endian required)", executable);
        return false;
    }

    if (executable->header.identity[7] != ELF_SYSABI_SYSTEM_V) {
        log(LOG_ERROR, "ELF <0x%x> is invalid (SystemV SysABI required)", executable);
        return false;
    }

    if (executable->header.isa != ELF_ISA_X86_64) {
        log(LOG_ERROR, "ELF <0x%x> is invalid (x86_64 ISA required)", executable);
        return false;
    }

    return true;
}

void elf_executable_extract_symbols(elf_executable_temp_T* executable_temp) {
    uint64_t num_symbols = executable_temp->symbol_table->length / executable_temp->symbol_table->entry_size;
    executable_temp->executable->symbol_table = symbol_table_init(num_symbols);

    for (int i = 0; i < num_symbols; i++) {
        symbol_T            symbol;
        elf_symbol_T*       symbol_elf = &((elf_symbol_T*)&executable_temp->buffer[executable_temp->symbol_table->offset])[i];
        elf_symbol_type_E   type       = ELF_SYMBOL_TYPE(symbol_elf->info);

        symbol.name    = &executable_temp->executable->string_table[symbol_elf->name_offset];
        symbol.address = symbol_elf->value;

        switch (type) {
            case ELF_SYMBOL_FUNC: {
                symbol.type = SYMBOL_FUNCTION;
                break;
            }
            case ELF_SYMBOL_OBJECT: {
                symbol.type = SYMBOL_VARIABLE;
                break;
            }
            default: {
                symbol.type = SYMBOL_UNKNOWN;
                break;
            }
        }

        symbol_table_insert(&executable_temp->executable->symbol_table, symbol);
    }
}

elf_executable_T* elf_executable_create(uint8_t* buffer) {
    elf_executable_temp_T executable_temp;
    executable_temp.executable = memory_allocate(sizeof(elf_executable_T));
    memory_copy(buffer, &executable_temp.executable->header, sizeof(elf_header_T));

    if (!elf_executable_validate(executable_temp.executable)) {
        return NULL;
    }

//    log(LOG_INFO, "ELF header:");
//    log(LOG_INFO, "    -> Architecture: %s", g_elf_target_architecture_strings[executable_temp.executable->header.identity[4]]);
//    log(LOG_INFO, "    -> Endian: %s", g_elf_endianness_strings[executable_temp.executable->header.identity[5]]);
//    log(LOG_INFO, "    -> SysABI: %s", g_elf_sysabi_strings[executable_temp.executable->header.identity[7]]);
//    log(LOG_INFO, "    -> Type: %s", g_elf_object_type_strings[executable_temp.executable->header.type]);
//    log(LOG_INFO, "    -> ISA: %s", g_elf_instruction_set_strings[executable_temp.executable->header.isa]);
//    log(LOG_INFO, "    -> Entry: 0x%x", executable_temp.executable->header.address_entry_point);
//    log(LOG_INFO, "    -> PHDR offset: 0x%x", executable_temp.executable->header.offset_program_header);
//    log(LOG_INFO, "    -> PHDR num: %d", executable_temp.executable->header.num_program_header_entries);
//    log(LOG_INFO, "    -> SHDR offset: 0x%x", executable_temp.executable->header.offset_section_header);
//    log(LOG_INFO, "    -> SHDR num: %d\n", executable_temp.executable->header.num_section_header_entries);

    executable_temp.buffer = buffer;

//    log(LOG_INFO, "Section Headers:");
    elf_section_T* sections = (elf_section_T*)&buffer[executable_temp.executable->header.offset_section_header];

    executable_temp.section_header_string_table = &sections[executable_temp.executable->header.string_section_index];

    for (int i = 0; i < executable_temp.executable->header.num_section_header_entries; i++) {
        string_t name = (string_t)&(&buffer[executable_temp.section_header_string_table->offset])[sections[i].name_offset];
//        log(LOG_INFO, "    [%s] '%s'", g_elf_section_type_strings[MIN(sections[i].type, ELF_SECTION_ENUM_END)], name);

        switch (sections[i].type) {
            case ELF_SECTION_SYMBOL_TABLE: {
                executable_temp.symbol_table = &sections[i];
                break;
            }
            case ELF_SECTION_STRING_TABLE: {
                if (string_compare(name, ".strtab")) {
                    // copy the string table, so it stays accessible as long as the elf_executable_T exists
                    executable_temp.executable->string_table = memory_allocate(sections[i].length);
                    memory_copy(&buffer[sections[i].offset], executable_temp.executable->string_table, sections[i].length);
                }
                break;
            }
        }
    }

//    log(LOG_INFO, "Program Headers:");
    executable_temp.executable->num_mappings = 0;

    elf_segment_T* segments = (elf_segment_T*)&buffer[executable_temp.executable->header.offset_program_header];
    for (int i = 0; i < executable_temp.executable->header.num_program_header_entries; i++) {
//        log(LOG_INFO, "    %s", g_elf_segment_type_strings[MIN(segments[i].type, ELF_SEGMENT_ENUM_END)]);

        switch (segments[i].type) {
            case ELF_SEGMENT_LOAD: {
                executable_temp.executable->num_mappings++;
                break;
            }
        }
    }

    executable_temp.executable->mappings = memory_allocate(executable_temp.executable->num_mappings * sizeof(elf_mapping_T));

    uint64_t mappings_index = 0;
    segments = (elf_segment_T*)&buffer[executable_temp.executable->header.offset_program_header];
    for (int i = 0; i < executable_temp.executable->header.num_program_header_entries; i++) {
        switch (segments[i].type) {
            case ELF_SEGMENT_LOAD: {
                executable_temp.executable->mappings[mappings_index] = (elf_mapping_T){
                    .offset_file    = segments[i].offset,
                    .offset_virtual = segments[i].address_virtual,
                    .length_file    = segments[i].length_file,
                    .length_virtual = segments[i].length_memory
                };
                mappings_index++;
                break;
            }
        }
    }

    elf_executable_extract_symbols(&executable_temp);

    return executable_temp.executable;
}

void elf_executable_destruct(elf_executable_T* executable) {
    memory_free(executable->string_table);
    symbol_table_destruct(&executable->symbol_table);
    memory_free(executable->mappings);
    memory_free(executable);
}