// This file is part of noxos and licensed under the MIT open source license

#include "drivers/builtin/fs/vfs.h"
#include "drivers/builtin/fs/ramfs.h"
#include "drivers/builtin/fs/ustar.h"
#include "utils/memory.h"
#include "utils/logger.h"
#include "utils/math.h"

fs_T g_root_fs;
fs_T g_ramdisk_fs;

string_t g_fs_names[FS_ENUM_END] = {
        "RAMFS"
};

vfs_node_cache_T* vfs_node_cache_create(vfs_node_T* node, uint64_t size) {
    vfs_node_cache_T* node_cache = memory_allocate(sizeof(vfs_node_cache_T));
    node_cache->node        = node;
    node_cache->buffer_size = size;
    node_cache->buffer      = memory_allocate(size);
    node_cache->reclaimable = false;
    node->cache             = node_cache;
    return node_cache;
}

void vfs_node_cache_destruct(vfs_node_cache_T* node_cache) {
    node_cache->node->cache = NULL;
    memory_free(node_cache->buffer);
    memory_free(node_cache);
}

void vfs_node_bring_to_parent_start(vfs_node_T* node) {
    if (node->parent == NULL) { return; }

    if (node->parent->childs == node) {
        return;
    }

    if (node->parent->childs == NULL) {
        node->parent->childs = node;
        return;
    }

    if (node->prev != NULL) {
        node->prev->next = node->next;
    }

    if (node->next != NULL) {
        node->next->prev = node->prev;
    }

    node->parent->childs->prev = node;
    node->next                 = node->parent->childs;
    node->parent->childs       = node;
}

vfs_node_T* vfs_node_create(vfs_node_T* parent, string_t name, vfs_node_type_E type, void* specific) {
    fs_T* filesystem = NULL;

    if (parent != NULL) {
        filesystem = parent->filesystem;

        switch (parent->type) {
            case VFS_NODE_DIRECTORY: {
                break;
            }
            case VFS_NODE_BLOCK_DEVICE: {
                log(LOG_WARNING, "<VFS> failed to create node (block device cannot have childs)");
                return NULL;
            }
            case VFS_NODE_FILE: {
                log(LOG_WARNING, "<VFS> failed to create node (file cannot have childs)");
                return NULL;
            }
            case VFS_NODE_MOUNT: {
                parent = ((fs_T *) parent->specific)->root_node;
                break;
            }
            default: {
                // if this ever happens, something went really wrong
                log(LOG_ERROR, "<VFS> failed to create node (invalid parent type)");
                return NULL;
            }
        }
    }

    vfs_node_T* node = memory_allocate(sizeof(vfs_node_T));

    uint64_t strlen = string_length(name);
    if (strlen >= VFS_MAX_NAME_LENGTH) {
        log(LOG_WARNING, "<VFS> name '%s' is too long, shrinking to '%.*s'", name, VFS_MAX_NAME_LENGTH-1, name);
    }
    memory_copy((void*)name, node->name, VFS_MAX_NAME_LENGTH-1); node->name[VFS_MAX_NAME_LENGTH-1] = '\0';

    node->parent     = parent;
    node->type       = type;
    node->cache      = NULL;
    node->specific   = specific;
    node->filesystem = filesystem;
    node->childs     = NULL;
    node->prev       = NULL;
    node->next       = NULL;

    vfs_node_bring_to_parent_start(node);

    return node;
}

void vfs_node_dump_info(vfs_node_T* node, uint64_t indent) {
    char buf[indent+1];
    memory_set(buf, ' ', indent);
    buf[indent] = '\0';

    switch (node->type) {
        case VFS_NODE_DIRECTORY: {
            log(LOG_NONE, "%s[dir] %s", buf, node->name);
            vfs_node_T* child = node->childs;
            while (child != NULL) {
                vfs_node_dump_info(child, indent+2);
                child = child->next;
            }
            break;
        }
        case VFS_NODE_FILE: {
            log(LOG_NONE, "%s[file] %s", buf, node->name);
            break;
        }
        case VFS_NODE_MOUNT: {
            log(LOG_NONE, "%s[mount] %s -> %s", buf, node->name, g_fs_names[((fs_T*)node->specific)->type]);
            vfs_node_dump_info(((fs_T*)node->specific)->root_node, indent+6);
            break;
        }
    }
}

void vfs_node_destruct(vfs_node_T* node) {
    vfs_node_T* child = node->childs;
    while (child != NULL) {
        vfs_node_destruct(child);
        child = child->next;
    }

    if (node->parent->childs == node) {
        node->parent->childs = node->next;
    }

    if (node->prev != NULL) {
        node->prev->next = node->next;
    }

    if (node->next != NULL) {
        node->next->prev = node->prev;
    }

    memory_free(node);
}


void vfs_node_delete(vfs_node_T* node) {
    switch (node->type) {
        case VFS_NODE_FILE: {
            vfs_file_delete(node);
            break;
        }
        case VFS_NODE_DIRECTORY: {
            vfs_directory_delete(node);
        }
        default: {
            break;
        }
    }
}


vfs_node_T* vfs_node_resolve_child(vfs_node_T* node, string_t child_name) {
    vfs_node_T* child = node->childs;
    while (child != NULL) {
        if (!string_compare(child_name, child->name)) {
            child = child->next;
            continue;
        }

        vfs_node_bring_to_parent_start(child);
        return child;
    }

    return NULL;
}

vfs_node_T* vfs_file_create(fs_T* filesystem, string_t path) {
    // everything before index is the path and everything behind index is the name of the file that will be created
    uint64_t index = string_find_last(path, '/') + 1;

    char dir_path[index];
    memory_copy(path, dir_path, index); dir_path[index] = '\0';

    vfs_node_T* dir_node = vfs_resolve_path(filesystem, dir_path);
    if (dir_node == NULL) {
        log(LOG_WARNING, "<VFS> failed to create file %s (invalid path)", path);
        return NULL;
    }
    vfs_node_T* file = vfs_node_create(dir_node, &path[index], VFS_NODE_FILE, NULL);

    switch (file->filesystem->type) {
        default: {
            break;
        }
    }

    return file;
}

void vfs_file_delete(vfs_node_T* file) {
    switch (file->filesystem->type) {
        case FS_RAMFS: {
            ramfs_file_delete(file);
            break;
        }
        default: {
            break;
        }
    }
    vfs_node_destruct(file);
}

uint64_t vfs_file_write(vfs_node_T* file, uint64_t position, uint64_t size, uint8_t* buffer_in) {
    // in future this probably isn't a good way to determine the size of a file:
    file->size = size;

    switch (file->filesystem->type) {
        case FS_RAMFS: {
            return ramfs_file_write(file, size, buffer_in);
        }
        default: {
            break;
        }
    }

    return 0;
}

uint64_t vfs_file_read(vfs_node_T* file, uint64_t position, uint64_t size, uint8_t* buffer_out) {
    switch (file->filesystem->type) {
        case FS_RAMFS: {
            return ramfs_file_read(file, size, buffer_out);
        }
        default: {
            break;
        }
    }

    return 0;
}

vfs_node_T* vfs_directory_create(fs_T* filesystem, string_t path) {
    // everything before index is the path and everything behind index is the name of the dir that will be created
    uint64_t index = string_find_last(path, '/') + 1;

    char dir_path[index];
    memory_copy(path, dir_path, index); dir_path[index] = '\0';

    vfs_node_T* dir_node = vfs_resolve_path(filesystem, dir_path);
    if (dir_node == NULL) {
        log(LOG_WARNING, "<VFS> failed to create directory %s (invalid path)", path);
        return NULL;
    }
    vfs_node_T* dir = vfs_node_create(dir_node, &path[index], VFS_NODE_DIRECTORY, NULL);

    switch (dir->filesystem->type) {
        default: {
            break;
        }
    }

    return dir;
}

void vfs_directory_delete(vfs_node_T* directory) {
    vfs_node_destruct(directory);
}

void vfs_init(boot_info_T* boot_info) {
    g_root_fs.type                      = FS_RAMFS;
    g_root_fs.root_node                 = vfs_node_create(NULL, "root", VFS_NODE_DIRECTORY, NULL);
    g_root_fs.root_node->filesystem     = &g_root_fs;
    g_ramdisk_fs.type                   = FS_RAMFS;
    g_ramdisk_fs.root_node              = vfs_node_create(NULL, "root", VFS_NODE_DIRECTORY, NULL);
    g_ramdisk_fs.root_node->filesystem  = &g_ramdisk_fs;

    // unpacking initial ramdisk to '/initrd/'
    vfs_node_create(g_root_fs.root_node, "initrd", VFS_NODE_MOUNT, &g_ramdisk_fs);
    vfs_unpack_archive_ustar(&g_ramdisk_fs, boot_info->ramdisk_file->address);

    vfs_node_create(g_root_fs.root_node, "devices", VFS_NODE_DIRECTORY, NULL);
}

vfs_node_T* vfs_resolve_path(fs_T* filesystem, string_t path) {
    if (*path != '/') {
        log(LOG_WARNING, "<VFS> failed to resolve node from invalid path '%s' (no root prefix)", path);
        return NULL;
    }
    path++;

    uint32_t    length;
    vfs_node_T* node = filesystem->root_node;

    while(*path != '\0') {

        length = string_find_next(path, '/');

        char name[length+1];
        memory_copy((void*)path, name, length);
        name[length] = '\0';

        node = vfs_node_resolve_child(node, name);
        if (node == NULL) {
            log(LOG_WARNING, "<VFS> failed to resolve node from invalid path '%s' (node child resolve null)", path);
            return NULL;
        }
        path += length;

        switch (node->type) {
            case VFS_NODE_DIRECTORY: {
                path++; // remove the '/'
                break;
            }
            case VFS_NODE_MOUNT: {
                return vfs_resolve_path(node->specific, path);
            }
            default: {
                break;
            }
        }
    }

    return node;
}

void vfs_unpack_archive_ustar(fs_T* filesystem , void* archive) {
    // skip the first header, that specifies the root directory
    ustar_header_T* header = archive + sizeof(ustar_header_T);

    log(LOG_INFO, "<VFS> unpacking initrd");

    while (memory_compare(header->indicator, "ustar", 5)) {
        uint64_t size   = octal_string_to_int(header->size, 11);
        uint32_t blocks = 1 + (size > 0 ? (size / 512) + 1 : 0);
        if (header->type > 0) {
            header->type -= '0';
        }

        log(LOG_INFO, "    extracting: %s", header->name);

        switch (header->type) {
            case USTAR_FILE: {
                vfs_node_T* file = vfs_file_create(filesystem, &header->name[1]);
                vfs_file_write(file, 0, size, (uint8_t*)&header[1]);
                break;
            }
            case USTAR_DIRECTORY: {
                // get rid of the '/' at the end
                header->name[string_length(header->name)-1] = '\0';

                vfs_directory_create(filesystem, &header->name[1]);
                break;
            }
        }

        header = &header[blocks];
    }
}