// This file is part of noxos and licensed under the MIT open source license

#include "drivers/builtin/ps2/controller.h"
#include "drivers/builtin/ps2/keyboard.h"
#include "drivers/builtin/acpi/fadt.h"
#include "utils/io.h"
#include "utils/logger.h"
#include "boot/config.h"

void ps2_controller_init() {
    // check if the ps/2 controller exists
    if (g_sysconfig->ps2_acpi_validation) {
        if (!(g_acpi_table_fadt->ia_boot_architecture_flags & (1 << 1))) {
            log(LOG_ERROR, "<PS2> ps2 controller initialization failed (ps2 controller not set in fadt boot flags)");
            return;
        }
    } else {
        log(LOG_WARNING, "<PS2> skipping ps2 controller existence acpi check (disabled by sysconfig)");
    }

    // disable ps/2 devices
    ps2_controller_command(0xAD); // keyboard
    ps2_controller_command(0xA7); // mouse

    // clear buffer
    while (io_in_byte(PS2_CONTROLLER_STATUS_PORT) & 0b00000001) {
        io_in_byte(PS2_CONTROLLER_DATA_PORT);
    }

    // set the configuration byte
    uint8_t config_byte  = ps2_controller_command(0x20);
    bool    dual_channel = (config_byte & 0b00100000);
    config_byte &= 0b10111100;
    ps2_controller_command_with_data(0x60, config_byte);

    // performing self test
    if (ps2_controller_command(0xAA) != 0x55) {
        log(LOG_ERROR, "<PS2> controller driver init failed (controller self test failed)");
        return;
    }
    // restore the config byte, due to a possible reset by the self test
    ps2_controller_command_with_data(0x60, config_byte);

    // check if the controller supports 2 channels
    if (dual_channel) {
        ps2_controller_command(0xA8);
        dual_channel = !(ps2_controller_command(0x20) & 0b00100000);
        ps2_controller_command(0xA7);
    }

    // perform interface tests
    if (ps2_controller_command(0xAB) > 0 && (!dual_channel || ps2_controller_command(0xA9) > 0)) {
        log(LOG_ERROR, "<PS2> controller driver init failed (interface tests failed)");
    }

    // init devices
    ps2_keyboard_init();
}

uint8_t ps2_controller_command(uint8_t command) {
    io_out_byte(PS2_CONTROLLER_COMMAND_PORT, command);
    if (ps2_controller_command_has_response(command) && !ps2_controller_wait_until_ready_for_output()) {
        return 0;
    }
    return io_in_byte(PS2_CONTROLLER_DATA_PORT);
}

uint8_t ps2_controller_command_with_data(uint8_t command, uint8_t data) {
    io_out_byte(PS2_CONTROLLER_COMMAND_PORT, command);
    if (!ps2_controller_wait_until_ready_for_input()) {
        return 0;
    }
    io_out_byte(PS2_CONTROLLER_DATA_PORT, data);
    if (ps2_controller_command_has_response(command) && !ps2_controller_wait_until_ready_for_output()) {
        return 0;
    }
    return io_in_byte(PS2_CONTROLLER_DATA_PORT);
}

bool ps2_controller_command_has_response(uint8_t command) {
    if (command >= 0x20 && command <= 0x3F) return true;
    if (command >= 0xA9 && command <= 0xAC) return true;
    if (command == 0xC0)                    return true;
    if (command == 0xD0)                    return true;

    return false;
}

bool ps2_controller_wait_until_ready_for_input() {
    uint32_t timeout = 0;
    while (io_in_byte(PS2_CONTROLLER_STATUS_PORT) & 0b00000010) {
        if (timeout > PS2_CONTROLLER_TIMEOUT) {
            log(LOG_WARNING, "<PS2 Controller> Controller not ready for input (timeout)");
            return false;
        }
        timeout++;
    }
    return true;
}

bool ps2_controller_wait_until_ready_for_output() {
    uint32_t timeout = 0;
    while (!(io_in_byte(PS2_CONTROLLER_STATUS_PORT) & 0b00000001)) {
        if (timeout > PS2_CONTROLLER_TIMEOUT) {
            log(LOG_WARNING, "<PS2 Controller> Controller not ready for output (timeout)");
            return false;
        }
        timeout++;
    }
    return true;
}

uint8_t ps2_controller_read_data() {
    return io_in_byte(PS2_CONTROLLER_DATA_PORT);
}

void ps2_controller_write_data(uint8_t data) {
    io_out_byte(PS2_CONTROLLER_DATA_PORT, data);
}
