// This file is part of noxos and licensed under the MIT open source license

#include "drivers/builtin/ps2/keyboard.h"
#include "drivers/builtin/ps2/scancodes.h"
#include "drivers/builtin/ps2/controller.h"
#include "drivers/builtin/tty.h"
#include "utils/logger.h"
#include "utils/io.h"
#include "platform/interrupts.h"

char scancode_set_1[] = "  1234567890-=  qwertyuiop[]  asdfghjkl;'` \\zxcvbnm,./ *               789-456+1230.";
bool shift_enabled = false;
bool extended_key  = false;

uint8_t ps2_keyboard_command(ps2_keyboard_command_E command) {
    uint8_t response = PS2_KEYBOARD_RESPONSE_RESEND;
    uint8_t tries    = 0;

    while (response == PS2_KEYBOARD_RESPONSE_RESEND) {
        if (!ps2_controller_wait_until_ready_for_input()) {
            return 0;
        }
        ps2_controller_write_data(command);
        if (!ps2_controller_wait_until_ready_for_output()) {
            return 0;
        }
        while (io_in_byte(PS2_CONTROLLER_STATUS_PORT) & 0b00000001) {
            response = ps2_controller_read_data();
        }


        log(LOG_DEBUG, "<PS2> '%s'  ->  0x%xb", ps2_keyboard_command_to_string(command), response);
        tries++;

        if (tries >= 3) {
            log(LOG_WARNING, "<PS2> Command %s not supported by keyboard (to many resend responses)", ps2_keyboard_command_to_string(command));
            return 0;
        }
    }

    return response;
}

void ps2_keyboard_init() {

    // re-enable the ps/2 keyboard
    ps2_controller_command(0xAE);

    // reset the keyboard
    ps2_keyboard_command(PS2_KEYBOARD_COMMAND_SELF_TEST);

    // set scancode set to 1
    ps2_keyboard_command(PS2_KEYBOARD_COMMAND_SET_SCANCODE_SET);
    ps2_keyboard_command(1);

    // enable keyboard interrupt on ps/2 controller
    uint8_t config_byte = ps2_controller_command(0x20);
    config_byte |= 0b00000001;
    ps2_controller_command_with_data(0x60, config_byte);

    pic_unmask_irq(IRQ_KEYBOARD);
}

void ps2_keyboard_read() {
    uint8_t scancode = ps2_controller_read_data();
    char    chr [8];

    if (extended_key) {
        extended_key = false;

        switch (scancode) {
            case PS2_SCANCODE_SET_1_E0_CURSOR_LEFT: {
                uint8_t data = 0;
                string_add_ext_command(chr, STRING_EXT_CMD_CURSOR_MOVE, 1, &data);
                chr[STRING_EXT_SIZE(1)+1] = '\0';
                break;
            }
            case PS2_SCANCODE_SET_1_E0_CURSOR_RIGHT: {
                uint8_t data = 1;
                string_add_ext_command(chr, STRING_EXT_CMD_CURSOR_MOVE, 1, &data);
                chr[STRING_EXT_SIZE(1)+1] = '\0';
                break;
            }
            case PS2_SCANCODE_SET_1_E0_CURSOR_UP: {
                uint8_t data = 2;
                string_add_ext_command(chr, STRING_EXT_CMD_CURSOR_MOVE, 1, &data);
                chr[STRING_EXT_SIZE(1)+1] = '\0';
                break;
            }
            case PS2_SCANCODE_SET_1_E0_CURSOR_DOWN: {
                uint8_t data = 3;
                string_add_ext_command(chr, STRING_EXT_CMD_CURSOR_MOVE, 1, &data);
                chr[STRING_EXT_SIZE(1)+1] = '\0';
                break;
            }
            default: {
                return;
            }
        }

        tty_write(chr);
        return;
    }

    if (scancode == PS2_SCANCODE_SET_1_MODIFIER) {
        extended_key = true;
        return;
    }

    if (scancode >= PS2_SCANCODE_SET_1_RELEASE) {
        scancode -= PS2_SCANCODE_SET_1_RELEASE;
        switch (scancode) {
            case PS2_SCANCODE_SET_1_SHIFT_LEFT:
            case PS2_SCANCODE_SET_1_SHIFT_RIGHT: {
                shift_enabled = !shift_enabled;
                return;
            }
        }
        return;
    }

    switch (scancode) {
        case PS2_SCANCODE_SET_1_ESCAPE:
        case PS2_SCANCODE_SET_1_TABULATOR:
        case PS2_SCANCODE_SET_1_ALT_LEFT:
        case PS2_SCANCODE_SET_1_NUMLOCK:
        case PS2_SCANCODE_SET_1_SCROLLLOCK:
        case PS2_SCANCODE_SET_1_F1:
        case PS2_SCANCODE_SET_1_F2:
        case PS2_SCANCODE_SET_1_F3:
        case PS2_SCANCODE_SET_1_F4:
        case PS2_SCANCODE_SET_1_F5:
        case PS2_SCANCODE_SET_1_F6:
        case PS2_SCANCODE_SET_1_F7:
        case PS2_SCANCODE_SET_1_F8:
        case PS2_SCANCODE_SET_1_F9:
        case PS2_SCANCODE_SET_1_F10:
        case PS2_SCANCODE_SET_1_F11:
        case PS2_SCANCODE_SET_1_F12: {
            log(LOG_WARNING, "special keys aren't implemented yet");
            return;
        }

        case PS2_SCANCODE_SET_1_BACKSPACE: {
            chr[0] = '\b';
            chr[1] = '\0';
            break;
        }

        case PS2_SCANCODE_SET_1_SPACE: {
            chr[0] = ' ';
            chr[1] = '\0';
            break;
        }

        case PS2_SCANCODE_SET_1_RETURN: {
            chr[0] = '\n';
            chr[1] = '\0';
            break;
        }

        case PS2_SCANCODE_SET_1_SHIFT_LEFT:
        case PS2_SCANCODE_SET_1_SHIFT_RIGHT:
        case PS2_SCANCODE_SET_1_CAPSLOCK: {
            shift_enabled = !shift_enabled;
            return;
        }

        default: {
            chr[0] = scancode_set_1[scancode];
            if (shift_enabled) {
                if (string_is_char_lowercase(chr[0]))      chr[0] -= 32;
                else if (string_is_char_uppercase(chr[0])) chr[0] += 32;
            }
            chr[1] = '\0';
            break;
        }
    }

    tty_write(chr);
}

string_t ps2_keyboard_command_to_string(ps2_keyboard_command_E command) {
    switch (command) {
        case PS2_KEYBOARD_COMMAND_SET_LEDS: {
            return "Set LEDs";
        }
        case PS2_KEYBOARD_COMMAND_ECHO: {
            return "Echo";
        }
        case PS2_KEYBOARD_COMMAND_SET_SCANCODE_SET: {
            return "Get/Set Scancode Set";
        }
        case PS2_KEYBOARD_COMMAND_IDENTIFY: {
            return "Identify";
        }
        case PS2_KEYBOARD_COMMAND_SET_TYPEMATIC: {
            return "Set Typematic";
        }
        case PS2_KEYBOARD_COMMAND_ENABLE_SCAN: {
            return "Enable Scan";
        }
        case PS2_KEYBOARD_COMMAND_DISABLE_SCAN: {
            return "Disable Scan";
        }
        case PS2_KEYBOARD_COMMAND_SET_DEFAULT_PARAMS: {
            return "Set Default Parameter";
        }
        case PS2_KEYBOARD_COMMAND_RESEND_LAST_BYTE: {
            return "Resend Last Byte";
        }
        case PS2_KEYBOARD_COMMAND_SELF_TEST: {
            return "Self Test";
        }
        default: {
            return "Invalid Command";
        }
    }
}