// This file is part of noxos and licensed under the MIT open source license

#include <drivers/driver.h>
#include <drivers/builtin/elf/relocation.h>
#include <drivers/builtin/json.h>
#include <utils/logger.h>
#include <utils/memory.h>
#include <proc/scheduler.h>
#include <platform/syscall.h>
#include <mm/region.h>

driver_manager_T g_driver_manager;

void driver_manager_init() {
    g_driver_manager.driver_pool         = driver_manager_chunk_alloc(NULL);
    g_driver_manager.lookup_table_pci    = hashmap_create(DRIVER_MANAGER_HASHMAP_SIZE);
    g_driver_manager.lookup_table_usb    = hashmap_create(DRIVER_MANAGER_HASHMAP_SIZE);
    g_driver_manager.lookup_table_fs_gpt = hashmap_create(DRIVER_MANAGER_HASHMAP_SIZE);
    g_driver_manager.lookup_table_fs_mbr = hashmap_create(DRIVER_MANAGER_HASHMAP_SIZE);

    vfs_node_T* node = vfs_resolve_path(&g_root_fs, "/initrd/drivers.json");
    if (node == NULL || node->type != VFS_NODE_FILE) {
        log(LOG_WARNING, "No drivers config file found. !!! THIS WILL VERY LIKELY CAUSE THE OS TO NOT WORK!!!");
        return;
    }

    char* buffer = memory_allocate(node->size + 1);
    vfs_file_read(node, 0, node->size, buffer);
    buffer[node->size] = 0;

    json_T* json = json_from_string(buffer);

    json_node_T* child = json->root_node->childs_start;
    while (child != NULL) {
        if (string_compare("BOOT", child->string)) {
            if (&child->childs_start[0] == NULL || child->childs_start[0].type != JSON_NODE_ARRAY) {
                log(LOG_WARNING, "sysconfig -> failed to parse field 'BOOT' (expected array)");
            } else {
                log(LOG_INFO, "Loading drivers for level 'BOOT':");
                json_node_T* array_child = child->childs_start[0].childs_start;
                while (array_child != NULL) {
                    uint64_t drv_id, fd;
                    log(LOG_INFO, "    > %s", array_child->string);
                    syscall_perform(SYSCALL_NX_FS_OPEN, (uint64_t)array_child->string, 0, (uint64_t)&fd, 0);
                    syscall_perform(SYSCALL_NX_DRV_REGISTER, fd, (uint64_t)&drv_id, 0, 0);
                    syscall_perform(SYSCALL_NX_FS_CLOSE, fd, 0, 0, 0);
                    array_child = array_child->next;
                }
            }
        } else {
            log(LOG_WARNING, "sysconfig -> skipping field '%s' (unknown identifier)", child->string);
        }

        child = child->next;
    }


    json_destruct(json);

    memory_free(buffer);
}

driver_manager_chunk_T* driver_manager_chunk_alloc(driver_manager_chunk_T* prev) {
    driver_manager_chunk_T* chunk = memory_allocate(sizeof(driver_manager_chunk_T));

    chunk->drivers_bitmap = bitmap_init(DRIVER_MANAGER_CHUNK_SIZE);
    chunk->free_slots     = DRIVER_MANAGER_CHUNK_SIZE;
    chunk->prev           = prev;

    if (chunk->prev != NULL) {
        chunk->prev->next = chunk;
    }

    return chunk;
}

driver_T* driver_find_pci_device(uint16_t vendor_id, uint16_t device_id) {
    log(LOG_WARNING, "driver manager -> pci driver lookup on disk is not implemented yet");
    return NULL;
}

driver_T* driver_find_usb_device(uint16_t vendor_id, uint16_t device_id) {
    log(LOG_WARNING, "driver manager -> usb driver lookup on disk is not implemented yet");
    return NULL;
}

driver_T* driver_find_fs_gpt(uint8_t guid[16]) {
    log(LOG_WARNING, "driver manager -> gpt driver lookup on disk is not implemented yet");
    return NULL;
}

driver_T* driver_find_fs_mbr(uint8_t type) {
    log(LOG_WARNING, "driver manager -> mbr driver lookup on disk is not implemented yet");
    return NULL;
}

void driver_init_pci(driver_T* driver, driver_config_pci_T* conf) {
    for (int i = 0; i < conf->num_device_ids; i++) {
        driver_config_device_id_T* device_id = &conf->device_ids[i];

        hashmap_T* level_2 = hashmap_lookup(&g_driver_manager.lookup_table_pci, device_id->vendor_id);
        if (level_2 == NULL) {
            level_2 = memory_allocate(sizeof(hashmap_T));
            *level_2 = hashmap_create(DRIVER_MANAGER_HASHMAP_SIZE);
            hashmap_insert(&g_driver_manager.lookup_table_pci, device_id->vendor_id, level_2);
        }

        void* entry = hashmap_lookup(level_2, device_id->device_id);
        if (entry != NULL) {
            log(LOG_WARNING, "Failed to register driver for pci device %xw:%xw (there is already a driver registered)", device_id->vendor_id, device_id->device_id);
            continue;
        }
        hashmap_insert(level_2, device_id->device_id, driver);
        log(LOG_INFO, "Registered driver for pci device %xw:%xw", device_id->vendor_id, device_id->device_id);
    }
}

void driver_init_usb(driver_T* driver, driver_config_usb_T* conf) {
    for (int i = 0; i < conf->num_device_ids; i++) {
        driver_config_device_id_T* device_id = &conf->device_ids[i];

        hashmap_T* level_2 = hashmap_lookup(&g_driver_manager.lookup_table_usb, device_id->vendor_id);
        if (level_2 == NULL) {
            level_2 = memory_allocate(sizeof(hashmap_T));
            *level_2 = hashmap_create(DRIVER_MANAGER_HASHMAP_SIZE);
            hashmap_insert(&g_driver_manager.lookup_table_usb, device_id->vendor_id, level_2);
        }

        void* entry = hashmap_lookup(level_2, device_id->device_id);
        if (entry != NULL) {
            log(LOG_WARNING, "Failed to register driver for usb device %xw:%xw (there is already a driver registered)", device_id->vendor_id, device_id->device_id);
            continue;
        }
        hashmap_insert(level_2, device_id->device_id, driver);
        log(LOG_INFO, "Registered driver for usb device %xw:%xw", device_id->vendor_id, device_id->device_id);
    }
}

void driver_init_fs_gpt(driver_T* driver, uint8_t gpt_guid[16]) {
    uint64_t  gpt_hash = ((uint64_t*)gpt_guid)[0] ^ ((uint64_t*)gpt_guid)[1];

    if (hashmap_lookup(&g_driver_manager.lookup_table_fs_gpt, gpt_hash) != NULL) {
        log(LOG_WARNING, "Failed to register driver for gpt file system: %xb%xb%xb%xb-%xb%xb-%xb%xb-%xb%xb-%xb%xb%xb%xb%xb%xb (there is already a driver registered)",
            gpt_guid[0x0], gpt_guid[0x1], gpt_guid[0x2], gpt_guid[0x3],
            gpt_guid[0x4], gpt_guid[0x5], gpt_guid[0x6], gpt_guid[0x7],
            gpt_guid[0x8], gpt_guid[0x9], gpt_guid[0xA], gpt_guid[0xB],
            gpt_guid[0xC], gpt_guid[0xD], gpt_guid[0xE], gpt_guid[0xF]);
        return;
    }

    hashmap_insert(&g_driver_manager.lookup_table_fs_gpt, gpt_hash, driver);

    log(LOG_INFO, "Registered driver for gpt file system: %xb%xb%xb%xb-%xb%xb-%xb%xb-%xb%xb-%xb%xb%xb%xb%xb%xb",
        gpt_guid[0x0], gpt_guid[0x1], gpt_guid[0x2], gpt_guid[0x3],
        gpt_guid[0x4], gpt_guid[0x5], gpt_guid[0x6], gpt_guid[0x7],
        gpt_guid[0x8], gpt_guid[0x9], gpt_guid[0xA], gpt_guid[0xB],
        gpt_guid[0xC], gpt_guid[0xD], gpt_guid[0xE], gpt_guid[0xF]);
}

void driver_init_fs_mbr(driver_T* driver, uint8_t type) {
    if (hashmap_lookup(&g_driver_manager.lookup_table_fs_gpt, type) != NULL) {
        log(LOG_WARNING, "Failed to register driver for mbr file system: %xb (there is already a driver registered)", type);
        return;
    }

    hashmap_insert(&g_driver_manager.lookup_table_fs_gpt, type, driver);

    log(LOG_INFO, "Registered driver for mbr file system: %xb", type);
}

void driver_init_fs(driver_T* driver, driver_config_fs_T* conf) {
    driver_init_fs_gpt(driver, conf->gpt_guid);
    driver_init_fs_mbr(driver, conf->mbr_type);
}

void driver_init(driver_T* driver, driver_transport_protocol_E protocol, conf_U conf) {
    switch (protocol) {
        case DRIVER_TRANSPORT_PCI: {
            driver_init_pci(driver, conf.pci);
            break;
        }
        case DRIVER_TRANSPORT_USB: {
            driver_init_usb(driver, conf.usb);
            break;
        }
        case DRIVER_TRANSPORT_FS: {
            driver_init_fs(driver, conf.fs);
            break;
        }
    }
}

driver_T* driver_register(elf_executable_T* executable, uint8_t* buffer) {
    uint64_t chunk_id = 0;
    driver_manager_chunk_T* chunk = g_driver_manager.driver_pool;
    while (chunk->free_slots == 0) {
        if (chunk->next == NULL) {
            driver_manager_chunk_alloc(chunk);
        }
        chunk = chunk->next;
        chunk_id++;
    }

    driver_T* driver;
    for (uint64_t i = 0; i < DRIVER_MANAGER_CHUNK_SIZE; i++) {
        if (bitmap_get(&chunk->drivers_bitmap, i)) continue;

        bitmap_set(&chunk->drivers_bitmap, i, true);
        chunk->free_slots -= 1;
        driver = &chunk->drivers[i];
        driver->id = (chunk_id * DRIVER_MANAGER_CHUNK_SIZE) + i;
        break;
    }

    driver->base       = (void*)(MEM_REGION_KERNEL_DRIVERS + (driver->id * DRIVER_MEMORY_REGION_SIZE));
    driver->executable = executable;


    elf_mappings_apply(executable->mappings, executable->num_mappings, buffer, driver->base, g_kernel_page_map);
    bool status = elf_relocate_driver(executable, buffer, driver->base);
    if (!status) {
        bitmap_set(&chunk->drivers_bitmap, driver->id % DRIVER_MANAGER_CHUNK_SIZE, false);
        chunk->free_slots += 1;
        return NULL;
    }


    return driver;
}

driver_T* driver_lookup_pci_device(uint16_t vendor_id, uint16_t device_id) {
    hashmap_T* level_2 = hashmap_lookup(&g_driver_manager.lookup_table_pci, vendor_id);
    if (level_2 == NULL) return driver_find_pci_device(vendor_id, device_id);

    driver_T* driver = hashmap_lookup(level_2, device_id);
    if (driver == NULL) return driver_find_pci_device(vendor_id, device_id);

    return driver;
}

driver_T* driver_lookup_usb_device(uint16_t vendor_id, uint16_t device_id) {
    hashmap_T* level_2 = hashmap_lookup(&g_driver_manager.lookup_table_usb, vendor_id);
    if (level_2 == NULL) return driver_find_usb_device(vendor_id, device_id);

    driver_T* driver = hashmap_lookup(level_2, device_id);
    if (driver == NULL) return driver_find_usb_device(vendor_id, device_id);

    return driver;
}

driver_T* driver_lookup_fs_gpt(uint8_t guid[16]) {
    uint64_t  gpt_hash = ((uint64_t*)guid)[0] ^ ((uint64_t*)guid)[1];

    driver_T* driver = hashmap_lookup(&g_driver_manager.lookup_table_fs_gpt, gpt_hash);
    if (driver == NULL) return driver_find_fs_gpt(guid);

    return driver;
}

driver_T* driver_lookup_fs_mbr(uint8_t type) {
    driver_T* driver = hashmap_lookup(&g_driver_manager.lookup_table_fs_gpt, type);
    if (driver == NULL) return driver_find_fs_mbr(type);

    return driver;
}

driver_T* driver_lookup_by_address(void* address) {
    uint64_t value = (uint64_t)address;
    value  &= 0xFFFFFFFF00000000;
    value  -= MEM_REGION_KERNEL_DRIVERS;
    value >>= 32;

    uint64_t chunk_id = value / DRIVER_MANAGER_CHUNK_SIZE;
    uint64_t local_id = value % DRIVER_MANAGER_CHUNK_SIZE;

    driver_manager_chunk_T* chunk = g_driver_manager.driver_pool;
    for (uint64_t i = 0; i < chunk_id; i++) {
        if (chunk->next == NULL) return NULL;
        chunk = chunk->next;
    }

    if (!bitmap_get(&chunk->drivers_bitmap, local_id)) return NULL;

    return &chunk->drivers[local_id];
}