// This file is part of noxos and licensed under the MIT open source license

#include "mm/heap.h"
#include "mm/page_map.h"
#include "mm/page_frame.h"
#include "utils/math.h"
#include "utils/logger.h"
#include "utils/memory.h"

void heap_segment_init(heap_segment_T* segment, heap_segment_T* next, heap_segment_T* prev, uint64_t size) {
    segment->magic  = HEAP_SEGMENT_MAGIC;
    segment->next   = next;
    segment->prev   = prev;
    segment->size   = size;
    segment->free   = true;
}

bool heap_segment_validate(heap_segment_T* segment) {
    return segment->magic == HEAP_SEGMENT_MAGIC;
}

bool heap_segment_combine_prev(heap_T* heap, heap_segment_T* segment) {
    if (segment->prev == NULL) { return false; }
    if (!segment->free || !segment->prev->free) { return false; }

    // link around the segment
    segment->prev->next = segment->next;
    if (segment->next != NULL) {
        segment->next->prev = segment->prev;
    }

    if (segment == heap->last_segment) {
        heap->last_segment = segment->prev;
    }

    // fuse with previous segment
    segment->prev->size += sizeof(heap_segment_T) + segment->size;

    // invalidate the segment
    segment->magic = 0;
    return true;
}

bool heap_segment_combine_next(heap_T* heap, heap_segment_T* segment) {
    if (segment->next == NULL) { return false; }

    return heap_segment_combine_prev(heap, segment->next);
}

bool heap_segment_shrink(heap_T* heap, heap_segment_T* segment, uint64_t size) {
    if (segment->size - size <= sizeof(heap_segment_T) + 0x10) { return false; }

    // break out remainding part into its own segment
    heap_segment_T* new_segment = (heap_segment_T*)((uint64_t)segment + sizeof(heap_segment_T) + size);
    heap_segment_init(new_segment, segment->next, segment->prev, segment->size - size - sizeof(heap_segment_T));

    segment->size = size;

    // bind new segment into chain
    segment->next = new_segment;
    if (new_segment->next != NULL) {
        new_segment->next->prev = new_segment;
    }
    new_segment->prev = segment;

    // try to combine the new segment with the segment that comes after it
    heap_segment_combine_next(heap, new_segment);

    if (segment == heap->last_segment) {
        heap->last_segment = new_segment;
    }

    return true;
}

void heap_init(heap_T* heap, void* base) {
    uint64_t num_pages = 2;
    uint64_t size      = num_pages * PFRAME_SIZE;

    for (int i = 0; i < num_pages; i++) {
        page_map_map_memory(g_kernel_page_map, &base[i * PFRAME_SIZE], pframe_request(), PM_FLAG_READ_WRITE);
    }

    heap_segment_T* base_segment = base;

    heap_segment_init(base_segment, NULL, NULL, size - sizeof(heap_segment_T));

    heap->start        = base;
    heap->end          = base + size;
    heap->last_segment = base_segment;
}

heap_segment_T* heap_append_segment(heap_T* heap, uint64_t minimum_size) {
    uint64_t size = CEIL_TO((minimum_size + sizeof(heap_segment_T)), PFRAME_SIZE);

    uint32_t num_pages = size / PFRAME_SIZE;
    for (uint32_t i = 0; i < num_pages; i++) {
        page_map_map_memory(g_kernel_page_map, heap->end, pframe_request(), PM_FLAG_READ_WRITE);
        heap->end += PFRAME_SIZE;
    }

    heap_segment_T* segment = heap->end - size;
    heap_segment_init(segment, NULL, heap->last_segment, size - sizeof(heap_segment_T));
    heap->last_segment->next = segment;

    if (heap_segment_combine_prev(heap, segment)) {
        segment = heap->last_segment;
    } else {
        heap->last_segment = segment;
    }

    return segment;
}

void* heap_memory_allocate(heap_T* heap, uint64_t size) {
    if (heap->start == NULL) { return NULL; }

    size = CEIL_TO(size, 0x10);

    // search for free segments, that are big enough
    heap_segment_T* segment = heap->start;
    while (segment != NULL) {
        if (!heap_segment_validate(segment)) {
            log(LOG_ERROR, "Heap Segment (0x%x) Invalid! while trying to allocate memory", segment);
            heap_dump_segments(heap);
            return NULL;
        }

        if (!segment->free || segment->size < size) {
            segment = segment->next;
            continue;
        }

        if (segment->size > size) {
            heap_segment_shrink(heap, segment, size);
        }

        segment->free = false;
        return (void*)((uint64_t)segment + sizeof(heap_segment_T));
    }

    segment = heap_append_segment(heap, size);
    if (segment->size > size) {
        heap_segment_shrink(heap, segment, size);
    }
    segment->free = false;

    return (void*)((uint64_t)segment + sizeof(heap_segment_T));
}

void heap_dump_segments(heap_T* heap) {
    log(LOG_DEBUG, "<=====[ Dumping Heap Segments ]=====>");

    log(LOG_NONE, "    |--------------------------|");
    log(LOG_NONE, "    |           START          |");
    log(LOG_NONE, "    | 0x%x       |", heap->start);
    log(LOG_NONE, "    |-------------v------------|");

    heap_segment_T* segment = heap->start;
    while (segment != NULL) {
        if (segment->magic != HEAP_SEGMENT_MAGIC) {
            log(LOG_NONE, "                  |");
            log(LOG_NONE, "                  |");
            log(LOG_NONE, "    !~~~~~~~~~~~~~^~~~~~~~~~~~~!");
            log(LOG_NONE, "    !  !!! Heap corrupted !!!  !");
            log(LOG_NONE, "    !~~~~~~~~~~~~~~~~~~~~~~~~~~!");
            memory_hexdump(segment, sizeof(heap_segment_T));
            return;
        }

        log(LOG_NONE, "                  |");
        log(LOG_NONE, "                  |");
        log(LOG_NONE, "    |-------------^------------|");
        log(LOG_NONE, "    | Header (size: 0x%xw)    |", sizeof(heap_segment_T));
        log(LOG_NONE, "    |--------------------------|");
        log(LOG_NONE, "    | 0x%x       |", (uint64_t)segment);
        log(LOG_NONE, "    | Valid: %?              |", heap_segment_validate(segment));
        log(LOG_NONE, "    | Free: %?              |", segment->free);
        log(LOG_NONE, "    | Size: 0x%xd         |", segment->size);
        log(LOG_NONE, "    | Prev: 0x%x |", (uint64_t)segment->prev);
        log(LOG_NONE, "    | Next: 0x%x |", (uint64_t)segment->next);
        log(LOG_NONE, "    |-------------v------------|");

        segment = segment->next;
    }

    log(LOG_NONE, "                  |");
    log(LOG_NONE, "                  |");
    log(LOG_NONE, "    |-------------^------------|");
    log(LOG_NONE, "    |            END           |");
    log(LOG_NONE, "    | 0x%x       |", heap->end);
    log(LOG_NONE, "    |--------------------------|");
}

void heap_memory_free(heap_T* heap, void* address) {
    if (heap->start == NULL) { return; }

    heap_segment_T* segment = (heap_segment_T*)(address - sizeof(heap_segment_T));
    if (!heap_segment_validate(segment)) {
        log(LOG_ERROR, "Heap Segment Invalid! while trying to free 0x%x", address);
        return;
    }
    segment->free = true;

    // try to combine in both directions, so we don't splitter the heap in a ton of small segments
    heap_segment_combine_prev(heap, segment);
    heap_segment_combine_next(heap, segment);

}

void heap_destruct(heap_T* heap) {
    log(LOG_INFO, "Destructing heap 0x%x", heap);

    heap_segment_T* segment = heap->start;
    while (segment != NULL) {
        segment->magic = 0;
        log(LOG_INFO, "    > Invalidated segment 0x%x + 0x%xw", segment, segment->size);
        segment = segment->next;
    }

    for (uint64_t i = (uint64_t)heap->start; i < (uint64_t)heap->end; i += PFRAME_SIZE) {
        void* pframe = page_map_get_physical_address(g_kernel_page_map, (void*)i);
        pframe_free(pframe);
        log(LOG_INFO, "    > Freed pframe 0x%xd", pframe);
        page_map_unmap_memory(g_kernel_page_map, (void*)i);
        log(LOG_INFO, "    > Unmapped 0x%x", i);
    }
}
