// This file is part of noxos and licensed under the MIT open source license

#include "mm/page_frame.h"
#include "mm/memory_map.h"
#include "utils/logger.h"
#include "utils/math.h"
#include "utils/core.h"

page_frame_manager_T g_page_frame_manager;

void pframe_reserve(void* address) {
    CORE_INTERRUPTABLE_HALT_WHILE(g_page_frame_manager.blocked)
    g_page_frame_manager.blocked = true;

    uint64_t index = (uint64_t)address / PFRAME_SIZE;

    if (bitmap_get(&g_page_frame_manager.page_bitmap, index)) {
        g_page_frame_manager.blocked = false;
        return;
    }

    if (bitmap_set(&g_page_frame_manager.page_bitmap, index, true)) {
        g_page_frame_manager.free_memory     -= PFRAME_SIZE;
        g_page_frame_manager.reserved_memory += PFRAME_SIZE;
    }

    g_page_frame_manager.blocked = false;
}

void pframe_reserve_multi(void* address, uint32_t n) {
    for (uint32_t i = 0; i < n; i++) {
        pframe_reserve(&address[i * PFRAME_SIZE]);
    }
}

void pframe_unreserve(void* address) {
    CORE_INTERRUPTABLE_HALT_WHILE(g_page_frame_manager.blocked)
    g_page_frame_manager.blocked = true;

    uint64_t index = (uint64_t)address / PFRAME_SIZE;

    if (!bitmap_get(&g_page_frame_manager.page_bitmap, index)) {
        g_page_frame_manager.blocked = false;
        return;
    }

    if (bitmap_set(&g_page_frame_manager.page_bitmap, index, false)) {
        g_page_frame_manager.free_memory     += PFRAME_SIZE;
        g_page_frame_manager.reserved_memory -= PFRAME_SIZE;

        if (g_page_frame_manager.page_bitmap_index > index) {
            g_page_frame_manager.page_bitmap_index = index;
        }
    }

    g_page_frame_manager.blocked = false;
}

void pframe_unreserve_multi(void* address, uint32_t n) {
    for (uint32_t i = 0; i < n; i++) {
        pframe_unreserve(&address[i * PFRAME_SIZE]);
    }
}

// this function should only be used by functions, that blocks the pframe manager
void pframe_lock(void* address) {
    uint64_t index = (uint64_t)address / PFRAME_SIZE;

    if (bitmap_get(&g_page_frame_manager.page_bitmap, index)) {
        return;
    }

    if (bitmap_set(&g_page_frame_manager.page_bitmap, index, true)) {
        g_page_frame_manager.free_memory -= PFRAME_SIZE;
        g_page_frame_manager.used_memory += PFRAME_SIZE;
    }
}

void pframe_lock_multi(void* address, uint32_t n) {
    for (uint32_t i = 0; i < n; i++) {
        pframe_lock(&address[i * PFRAME_SIZE]);
    }
}

void pframe_free(void* address) {
    CORE_INTERRUPTABLE_HALT_WHILE(g_page_frame_manager.blocked)
    g_page_frame_manager.blocked = true;

    uint64_t index = (uint64_t)address / PFRAME_SIZE;

    if (!bitmap_get(&g_page_frame_manager.page_bitmap, index)) {
        g_page_frame_manager.blocked = false;
        return;
    }

    if (bitmap_set(&g_page_frame_manager.page_bitmap, index, false)) {
        g_page_frame_manager.free_memory += PFRAME_SIZE;
        g_page_frame_manager.used_memory -= PFRAME_SIZE;

        if (g_page_frame_manager.page_bitmap_index > index) {
            g_page_frame_manager.page_bitmap_index = index;
        }
    }

    g_page_frame_manager.blocked = false;
}

void pframe_free_multi(void* address, uint32_t n) {
    for (uint32_t i = 0; i < n; i++) {
        pframe_free(&address[i * PFRAME_SIZE]);
    }
}

void* pframe_request() {
    CORE_INTERRUPTABLE_HALT_WHILE(g_page_frame_manager.blocked)
    g_page_frame_manager.blocked = true;

    while (g_page_frame_manager.page_bitmap_index < g_page_frame_manager.page_bitmap.size * 8) {
        if (bitmap_get(&g_page_frame_manager.page_bitmap, g_page_frame_manager.page_bitmap_index)) {
            g_page_frame_manager.page_bitmap_index++;
            continue;
        }
        void* address = (void*)(g_page_frame_manager.page_bitmap_index * PFRAME_SIZE);

        pframe_lock(address);

        g_page_frame_manager.blocked = false;
        return address;
    }

    g_page_frame_manager.blocked = false;
    return NULL;
}

void pframe_manager_init(boot_info_T* boot_info) {
    log(LOG_INFO, "<PageFrameManager> Initializing");

    struct limine_memmap_entry* largest_usable_entry = NULL;

    for (int i = 0; i < boot_info->memory_map->entry_count; i++) {
        struct limine_memmap_entry* entry = boot_info->memory_map->entries[i];

        if (entry->type != LIMINE_MEMMAP_USABLE) { continue; }
        if (largest_usable_entry != NULL && largest_usable_entry->length > entry->length) { continue; }

        largest_usable_entry = entry;
    }

    uint64_t total_memory = memory_map_get_total_memory_size(boot_info);
    log(LOG_INFO, "        * Total Memory size: %d B", total_memory);   // TODO: replace with format string

    g_page_frame_manager.page_bitmap       = bitmap_init_from_buffer((void*)largest_usable_entry->base, total_memory / PFRAME_SIZE / 8 + 1);
    g_page_frame_manager.page_bitmap_index = 0;
    g_page_frame_manager.reserved_memory   = 0;
    g_page_frame_manager.used_memory       = 0;
    g_page_frame_manager.free_memory       = total_memory;

    pframe_reserve_multi(0, total_memory / PFRAME_SIZE + 1);
    log(LOG_INFO, "        * Reserved all pages");

    for (int i = 0; i < boot_info->memory_map->entry_count; i++) {
        struct limine_memmap_entry* entry = boot_info->memory_map->entries[i];
        if (entry->type != LIMINE_MEMMAP_USABLE) { continue; }

        pframe_unreserve_multi((void*)entry->base, entry->length / PFRAME_SIZE);
    }
    log(LOG_INFO, "        * Unreserved usable pages");

    pframe_lock_multi(g_page_frame_manager.page_bitmap.buffer, g_page_frame_manager.page_bitmap.size / PFRAME_SIZE);
    log(LOG_INFO, "        * Locked page_bitmap pages");

    log(LOG_INFO, "        * Finished initialization");
    g_page_frame_manager.blocked = false;
}