#include "mm/page_map.h"
#include "mm/page_frame.h"
#include "mm/memory_map.h"
#include "mm/region.h"
#include "platform/interrupts.h"
#include "utils/memory.h"
#include "utils/math.h"
#include "utils/logger.h"

page_map_T* g_kernel_page_map;

page_map_T* page_map_create() {
    page_map_T* page_map = pframe_request();
    memory_set(page_map, 0, PFRAME_SIZE);

    // map the IDT into the new page map
    page_map_map_memory(page_map, g_idt_register.offset, g_idt_register.offset, 0);

    // map the kernel into the new page map
    uint64_t kernel_pages = CEIL_TO(KERNEL_END_ADDRESS - KERNEL_START_ADDRESS, PFRAME_SIZE) / PFRAME_SIZE;
    void*    address      = (void*)MEM_REGION_KERNEL_EXEC;

    for (uint64_t i = 0; i < kernel_pages; i++) {
        page_map_map_memory(page_map, address, page_map_get_physical_address(g_kernel_page_map, address), 0);
        address += PFRAME_SIZE;
    }

    return page_map;
}

page_map_T* page_map_request_entry(page_map_T* page_map, uint16_t index) {
    uint64_t* entry = &page_map->entries[index];

    if (page_map_entry_get_flag(entry, PM_FLAG_PRESENT)) {
        return page_map_entry_get_address(entry);
    }

    page_map_T* table = pframe_request();
    memory_set(table, 0, PFRAME_SIZE);
    page_map_entry_set_address(entry, (void*)((uint64_t)table >> 12));

    page_map_entry_set_flags(entry, PM_FLAG_PRESENT | PM_FLAG_READ_WRITE);

    return table;
}

void page_map_map_memory(page_map_T* page_map, void* virtual_address, void* physical_address, uint64_t flags) {
    virtual_address = (void*)FLOOR_TO((uint64_t)virtual_address, PFRAME_SIZE);

    // Intel's Developer Manual has a nice graph, that explains how these indexes are calculated
    uint16_t    page_index                  =  ((uint64_t)virtual_address >> 12) & 0x1FF;
    uint16_t    page_table_index            =  ((uint64_t)virtual_address >> 21) & 0x1FF;
    uint16_t    page_directory_index        =  ((uint64_t)virtual_address >> 30) & 0x1FF;
    uint16_t    page_directory_page_index   =  ((uint64_t)virtual_address >> 39) & 0x1FF;
    page_map_T* page_directory_page         =  page_map_request_entry(page_map, page_directory_page_index);
    page_map_T* page_directory              =  page_map_request_entry(page_directory_page, page_directory_index);
    page_map_T* page_table                  =  page_map_request_entry(page_directory, page_table_index);
    uint64_t*   page                        = &page_table->entries[page_index];

    *page = 0;
    page_map_entry_set_address(page, (void*)((uint64_t)physical_address >> 12));
    page_map_entry_set_flags(page, PM_FLAG_PRESENT | flags);
}

void page_map_unmap_memory(page_map_T* page_map, void* virtual_address) {
    virtual_address = (void*)FLOOR_TO((uint64_t)virtual_address, PFRAME_SIZE);

    // Intel's Developer Manual has a nice graph, that explains how these indexes are calculated
    uint16_t page_index                 =  ((uint64_t)virtual_address >> 12) & 0x1FF;
    uint16_t page_table_index           =  ((uint64_t)virtual_address >> 21) & 0x1FF;
    uint16_t page_directory_index       =  ((uint64_t)virtual_address >> 30) & 0x1FF;
    uint16_t page_directory_page_index  =  ((uint64_t)virtual_address >> 39) & 0x1FF;

    if (!page_map_entry_get_flag(&page_map->entries[page_directory_page_index], PM_FLAG_PRESENT)) {
        return;
    }
    page_map_T* page_directory_page = page_map_entry_get_address(&page_map->entries[page_directory_page_index]);

    if (!page_map_entry_get_flag(&page_directory_page->entries[page_directory_index], PM_FLAG_PRESENT)) {
        return;
    }
    page_map_T* page_directory = page_map_entry_get_address(&page_directory_page->entries[page_directory_index]);

    if (!page_map_entry_get_flag(&page_directory->entries[page_table_index], PM_FLAG_PRESENT)) {
        return;
    }
    page_map_T* page_table = page_map_entry_get_address(&page_directory->entries[page_table_index]);

    // zeroing out the complete entry
    page_table->entries[page_index] = 0;
}

bool page_map_check_memory(page_map_T* page_map, void* virtual_address) {
    virtual_address = (void*)FLOOR_TO((uint64_t)virtual_address, PFRAME_SIZE);

    // Intel's Developer Manual has a nice graph, that explains how these indexes are calculated
    uint16_t    page_index                  =  ((uint64_t)virtual_address >> 12) & 0x1FF;
    uint16_t    page_table_index            =  ((uint64_t)virtual_address >> 21) & 0x1FF;
    uint16_t    page_directory_index        =  ((uint64_t)virtual_address >> 30) & 0x1FF;
    uint16_t    page_directory_page_index   =  ((uint64_t)virtual_address >> 39) & 0x1FF;

    if (!page_map_entry_get_flag(&page_map->entries[page_directory_page_index], PM_FLAG_PRESENT)) return false;
    page_map_T* page_directory_page = page_map_entry_get_address(&page_map->entries[page_directory_page_index]);

    if (!page_map_entry_get_flag(&page_directory_page->entries[page_directory_index], PM_FLAG_PRESENT)) return false;
    page_map_T* page_directory = page_map_entry_get_address(&page_directory_page->entries[page_directory_index]);

    if (!page_map_entry_get_flag(&page_directory->entries[page_table_index], PM_FLAG_PRESENT)) return false;
    page_map_T* page_table = page_map_entry_get_address(&page_directory->entries[page_table_index]);

    return page_map_entry_get_flag(&page_table->entries[page_index], PM_FLAG_PRESENT);
}

void* page_map_get_physical_address(page_map_T* page_map, void* virtual_address) {
    virtual_address = (void*)FLOOR_TO((uint64_t)virtual_address, PFRAME_SIZE);

    // Intel's Developer Manual has a nice graph, that explains how these indexes are calculated
    uint16_t page_index                 =  ((uint64_t)virtual_address >> 12) & 0x1FF;
    uint16_t page_table_index           =  ((uint64_t)virtual_address >> 21) & 0x1FF;
    uint16_t page_directory_index       =  ((uint64_t)virtual_address >> 30) & 0x1FF;
    uint16_t page_directory_page_index  =  ((uint64_t)virtual_address >> 39) & 0x1FF;

    if (!page_map_entry_get_flag(&page_map->entries[page_directory_page_index], PM_FLAG_PRESENT)) {
        return NULL;
    }
    page_map_T* page_directory_page = page_map_entry_get_address(&page_map->entries[page_directory_page_index]);

    if (!page_map_entry_get_flag(&page_directory_page->entries[page_directory_index], PM_FLAG_PRESENT)) {
        return NULL;
    }
    page_map_T* page_directory = page_map_entry_get_address(&page_directory_page->entries[page_directory_index]);

    if (!page_map_entry_get_flag(&page_directory->entries[page_table_index], PM_FLAG_PRESENT)) {
        return NULL;
    }
    page_map_T* page_table = page_map_entry_get_address(&page_directory->entries[page_table_index]);

    if (!page_map_entry_get_flag(&page_table->entries[page_index], PM_FLAG_PRESENT)) {
        return NULL;
    }
    return page_map_entry_get_address(&page_table->entries[page_index]);
}

void page_map_destruct(page_map_T* page_map) {
    for (uint16_t page_directory_page_index = 0; page_directory_page_index < 512; page_directory_page_index++) {
        if (!page_map_entry_get_flag(&page_map->entries[page_directory_page_index], PM_FLAG_PRESENT)) {
            continue;
        }
        page_map_T* page_directory_page = page_map_entry_get_address(&page_map->entries[page_directory_page_index]);

        for (uint16_t page_directory_index = 0; page_directory_index < 512; page_directory_index++) {
            if (!page_map_entry_get_flag(&page_directory_page->entries[page_directory_index], PM_FLAG_PRESENT)) {
                continue;
            }
            page_map_T *page_directory = page_map_entry_get_address(&page_directory_page->entries[page_directory_index]);

            for (uint16_t page_table_index = 0; page_table_index < 512; page_table_index++) {
                if (!page_map_entry_get_flag(&page_directory->entries[page_table_index], PM_FLAG_PRESENT)) {
                    continue;
                }
                page_map_T *page_table = page_map_entry_get_address(&page_directory->entries[page_table_index]);


                pframe_free(page_table);
            }

            pframe_free(page_directory);
        }

        pframe_free(page_directory_page);
    }

    pframe_free(page_map);
}

void page_map_dump_info(page_map_T* page_map) {
    log(LOG_INFO, "Page map info:");
    log(LOG_INFO, "    Address: 0x%x", page_map);
    log(LOG_INFO, "    Mapped regions:");

    uint64_t mapped_pages        = 0;
    uint64_t num_pdp             = 0;
    uint64_t num_pd              = 0;
    uint64_t num_pt              = 0;
    bool     is_last_page_mapped = false;
    bool     in_lower_half       = true;
    uint64_t region_begin        = 0;

    log(LOG_INFO, "        Process Space:");

    for (uint16_t page_directory_page_index = 0; page_directory_page_index < 512; page_directory_page_index++) {
        if (!page_map_entry_get_flag(&page_map->entries[page_directory_page_index], PM_FLAG_PRESENT)) {
            continue;
        }
        page_map_T* page_directory_page = page_map_entry_get_address(&page_map->entries[page_directory_page_index]);

        for (uint16_t page_directory_index = 0; page_directory_index < 512; page_directory_index++) {
            if (!page_map_entry_get_flag(&page_directory_page->entries[page_directory_index], PM_FLAG_PRESENT)) {
                continue;
            }
            page_map_T *page_directory = page_map_entry_get_address(&page_directory_page->entries[page_directory_index]);

            for (uint16_t page_table_index = 0; page_table_index < 512; page_table_index++) {
                if (!page_map_entry_get_flag(&page_directory->entries[page_table_index], PM_FLAG_PRESENT)) {
                    continue;
                }
                page_map_T *page_table = page_map_entry_get_address(&page_directory->entries[page_table_index]);

                for (uint16_t page_index = 0; page_index < 512; page_index++) {
                    if (!page_map_entry_get_flag(&page_table->entries[page_index], PM_FLAG_PRESENT)) {
                        if (is_last_page_mapped) {

                            is_last_page_mapped = false;

                            uint64_t region_end = 0;

                            region_end  += page_directory_page_index;
                            region_end <<= 9;
                            region_end  += page_directory_index;
                            region_end <<= 9;
                            region_end  += page_table_index;
                            region_end <<= 9;
                            region_end  += page_index;
                            region_end <<= 12;

                            if (in_lower_half) {
                                if (region_end >= 0x0000800000000000) {
                                    in_lower_half = false;

                                    log(LOG_NONE, ""); // free line
                                    log(LOG_INFO, "        Kernel Space:");
                                }
                            }

                            log(LOG_INFO, "        0x%x  <-->  0x%x", region_begin, region_end);
                        }
                        continue;
                    }

                    if (!is_last_page_mapped) {
                        is_last_page_mapped = true;
                        region_begin = 0;

                        region_begin  += page_directory_page_index;
                        region_begin <<= 9;
                        region_begin  += page_directory_index;
                        region_begin <<= 9;
                        region_begin  += page_table_index;
                        region_begin <<= 9;
                        region_begin  += page_index;
                        region_begin <<= 12;
                    }

                    mapped_pages++;
                }

                num_pt++;
            }

            num_pd++;
        }

        num_pdp++;
    }

    log(LOG_INFO, "    Mapped pages: %d", mapped_pages);
    log(LOG_INFO, "    Mapped memory: %d KB", mapped_pages * 4);
    log(LOG_INFO, "    Structure Pages: %d", num_pdp + num_pd + num_pt);
    log(LOG_INFO, "        PML4 -> 1");
    log(LOG_INFO, "         |- PDP -> %d", num_pdp);
    log(LOG_INFO, "             |- PD -> %d", num_pd);
    log(LOG_INFO, "                 |- PT -> %d", num_pt);
}

void page_map_entry_set_flags(uint64_t* entry, uint64_t flags) {
    *entry |= flags;
}

bool page_map_entry_get_flag(uint64_t* entry, page_map_flag_E flag) {
    return ((*entry & flag) > 0);
}

void page_map_entry_set_address(uint64_t* entry, void* address) {
    uint64_t masked_address = (uint64_t)address & 0x000000ffffffffff;

    *entry &= 0xfff0000000000fff;
    *entry |= (masked_address << 12);
}

void* page_map_entry_get_address(uint64_t* entry) {
    return (void*)(*entry & 0x000ffffffffff000);
}

void paging_init() {
    // fetching the page map that was created by the bootloader
    g_kernel_page_map = page_map_fetch_current();

    log(LOG_INFO, "Paging Initialized");
}
