// This file is part of noxos and licensed under the MIT open source license

#include "platform/interrupts.h"
#include "platform/cpu.h"
#include "platform/gdt.h"
#include "platform/exceptions.h"
#include "platform/syscall.h"
#include "mm/page_frame.h"
#include "utils/logger.h"
#include "utils/io.h"
#include "proc/scheduler.h"
#include "drivers/builtin/ps2/keyboard.h"

idt_register_T g_idt_register;
uint8_t        g_handling_interrupt;

extern void idt_load(idt_register_T* idt_register);

// Exceptions
extern void interrupt_stub_0x00();
extern void interrupt_stub_0x01();
extern void interrupt_stub_0x02();
extern void interrupt_stub_0x03();
extern void interrupt_stub_0x04();
extern void interrupt_stub_0x05();
extern void interrupt_stub_0x06();
extern void interrupt_stub_0x07();
extern void interrupt_stub_0x08();
extern void interrupt_stub_0x09();
extern void interrupt_stub_0x0A();
extern void interrupt_stub_0x0B();
extern void interrupt_stub_0x0C();
extern void interrupt_stub_0x0D();
extern void interrupt_stub_0x0E();
extern void interrupt_stub_0x0F();
extern void interrupt_stub_0x10();
extern void interrupt_stub_0x11();
extern void interrupt_stub_0x12();
extern void interrupt_stub_0x13();
extern void interrupt_stub_0x14();
extern void interrupt_stub_0x15();
extern void interrupt_stub_0x16();
extern void interrupt_stub_0x17();
extern void interrupt_stub_0x18();
extern void interrupt_stub_0x19();
extern void interrupt_stub_0x1A();
extern void interrupt_stub_0x1B();
extern void interrupt_stub_0x1C();
extern void interrupt_stub_0x1D();
extern void interrupt_stub_0x1E();
extern void interrupt_stub_0x1F();

// IRQs on Master PIC
extern void interrupt_stub_0x20();
extern void interrupt_stub_0x21();
extern void interrupt_stub_0x22();
extern void interrupt_stub_0x23();
extern void interrupt_stub_0x24();
extern void interrupt_stub_0x25();
extern void interrupt_stub_0x26();
extern void interrupt_stub_0x27();

// IRQs on Slave PIC
extern void interrupt_stub_0x28();
extern void interrupt_stub_0x29();
extern void interrupt_stub_0x2A();
extern void interrupt_stub_0x2B();
extern void interrupt_stub_0x2C();
extern void interrupt_stub_0x2D();
extern void interrupt_stub_0x2E();
extern void interrupt_stub_0x2F();

// Custom
extern void interrupt_stub_0x80();

void idt_set_descriptor_entry_offset(idt_descriptor_entry_T* entry, uint64_t offset) {
    entry->offset0  = (uint16_t)  (offset & 0x000000000000ffff);
    entry->offset1  = (uint16_t) ((offset & 0x00000000ffff0000) >> 16);
    entry->offset2  = (uint32_t) ((offset & 0xffffffff00000000) >> 32);
}

uint64_t idt_get_descriptor_entry_offset(idt_descriptor_entry_T* entry) {
    uint64_t offset = 0;
    offset |= (uint64_t)entry->offset0;
    offset |= (uint64_t)entry->offset1 << 16;
    offset |= (uint64_t)entry->offset2 << 32;
    return offset;
}

void idt_set_gate(void* handler, uint8_t entry_offset, uint8_t type_attribute, uint8_t selector) {
    idt_descriptor_entry_T* entry = (idt_descriptor_entry_T*)(g_idt_register.offset + entry_offset * sizeof(idt_descriptor_entry_T));
    idt_set_descriptor_entry_offset(entry, (uint64_t)handler);
    entry->type_attribute = type_attribute;
    entry->selector       = selector;
}

void idt_init() {
    g_idt_register.limit   = 0x0FFF;
    g_idt_register.offset  = (uint64_t)pframe_request();

    // Exceptions
    idt_set_gate((void*)interrupt_stub_0x00, 0x00, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x01, 0x01, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x02, 0x02, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x03, 0x03, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x04, 0x04, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x05, 0x05, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x06, 0x06, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x07, 0x07, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x08, 0x08, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x09, 0x09, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x0A, 0x0A, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x0B, 0x0B, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x0C, 0x0C, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x0D, 0x0D, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x0E, 0x0E, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x0F, 0x0F, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x10, 0x10, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x11, 0x11, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x12, 0x12, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x13, 0x13, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x14, 0x14, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x15, 0x15, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x16, 0x16, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x17, 0x17, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x18, 0x18, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x19, 0x19, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x1A, 0x1A, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x1B, 0x1B, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x1C, 0x1C, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x1D, 0x1D, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x1E, 0x1E, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x1F, 0x1F, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);

    // IRQs on Master PIC
    idt_set_gate((void*)interrupt_stub_0x20, 0x20, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x21, 0x21, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x22, 0x22, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x23, 0x23, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x24, 0x24, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x25, 0x25, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x26, 0x26, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x27, 0x27, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);

    // IRQs on Slave PIC
    idt_set_gate((void*)interrupt_stub_0x28, 0x28, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x29, 0x29, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x2A, 0x2A, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x2B, 0x2B, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x2C, 0x2C, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x2D, 0x2D, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x2E, 0x2E, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);
    idt_set_gate((void*)interrupt_stub_0x2F, 0x2F, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);

    // Custom
    idt_set_gate((void*)interrupt_stub_0x80, 0x80, IDT_ENTRY_INTR_GATE, GDT_SELECTOR_KERNEL_CODE);

    idt_load(&g_idt_register);
    log(LOG_INFO, "Interrupts initialized");
}

void pic_init() {
    pic_remap();

    pic_unmask_irq(IRQ_PIT);
}

void pic_remap() {
    uint8_t master_masks = io_in_byte(PIC_MASTER_DATA); io_wait();
    uint8_t slave_masks  = io_in_byte(PIC_SLAVE_DATA);  io_wait();

    io_out_byte(PIC_MASTER_COMMAND, ICW1_INIT | ICW1_ICW4); io_wait();
    io_out_byte(PIC_SLAVE_COMMAND,  ICW1_INIT | ICW1_ICW4); io_wait();

    // set the IRQ offsets
    io_out_byte(PIC_MASTER_DATA, 0x20); io_wait();
    io_out_byte(PIC_SLAVE_DATA,  0x28); io_wait();

    io_out_byte(PIC_MASTER_DATA, 4); io_wait();
    io_out_byte(PIC_SLAVE_DATA,  2); io_wait();

    io_out_byte(PIC_MASTER_DATA, ICW4_8086); io_wait();
    io_out_byte(PIC_SLAVE_DATA,  ICW4_8086); io_wait();

    io_out_byte(PIC_MASTER_DATA, master_masks);
    io_out_byte(PIC_SLAVE_DATA,  slave_masks);
}

void pic_send_end_of_interrupt(pic_irq_E irq) {
    if (irq >= 8) {
        io_out_byte(PIC_SLAVE_COMMAND, PIC_END_OF_INTERRUPT);
    }
    io_out_byte(PIC_MASTER_COMMAND, PIC_END_OF_INTERRUPT);
}

void pic_mask_irq(pic_irq_E irq) {
    uint16_t port = PIC_MASTER_DATA;

    if (irq >= 8) {
        port = PIC_SLAVE_DATA;
        irq -= 8;
    }

    uint8_t value = io_in_byte(port) | (1 << irq);
    io_out_byte(port, value);
}

void pic_unmask_irq(pic_irq_E irq) {
    uint16_t port = PIC_MASTER_DATA;

    if (irq >= 8) {
        port = PIC_SLAVE_DATA;
        irq -= 8;
    }

    uint8_t value = io_in_byte(port) & ~(1 << irq);
    io_out_byte(port, value);
}

cpu_state_T* irq_handle(cpu_state_T* state, pic_irq_E irq) {
    cpu_state_T* return_state = state;

    switch (irq) {
        case IRQ_PIT: {
            return_state = scheduler_switch_context(state);
//            log(LOG_DEBUG, "PIT -> Tick");
            break;
        }
        case IRQ_KEYBOARD: {
            ps2_keyboard_read();
            break;
        }
        default: {
            log(LOG_WARNING, "Unhandled IRQ");
            break;
        }
    }

    pic_send_end_of_interrupt(irq);
    return return_state;
}

cpu_state_T* interrupts_handle_stage_2(cpu_state_T* state) {
    if (state->interrupt_id < EXCEPTIONS_ENUM_END) {
        return exception_handle(state);
    }

    if (state->interrupt_id < IRQ_MASTER_OFFSET + 0x10 && state->interrupt_id >= IRQ_MASTER_OFFSET) {
        return irq_handle(state, state->interrupt_id - IRQ_MASTER_OFFSET);
    }

    if (state->interrupt_id == 0x80) {
        return syscall_handle(state);
    }

    log(LOG_WARNING, "Unhandled interrupt: 0x%xb", state->interrupt_id);

    return state;
}

cpu_state_T* interrupts_handle(cpu_state_T* state) {
    g_handling_interrupt++;

    state = interrupts_handle_stage_2(state);

    g_handling_interrupt--;

    return state;
}

