// This file is part of noxos and licensed under the MIT open source license

#include "platform/syscall.h"
#include "utils/logger.h"
#include "utils/panic.h"
#include "utils/memory.h"
#include "proc/scheduler.h"
#include "drivers/builtin/fs/vfs.h"
#include "drivers/driver.h"
#include "mm/page_frame.h"
#include "mm/region.h"

void syscall_handle_nx_fs_open(cpu_state_T* state) {
    string_t           path   = (string_t)state->rdi;           // arg1
    uint64_t           len    = state->rsi;                     // arg2
    file_descriptor_t* fd_ptr = (file_descriptor_t*)state->rdx; // arg3

    vfs_node_T *node = vfs_resolve_path(&g_root_fs, path);
    if (node == NULL) {
        state->rax = STATUS_RESOURCE_NOT_AVAILABLE;
        return;
    }

    *fd_ptr = file_descriptor_request(scheduler_get_current_process()->fd_array, node);
}

void syscall_handle_nx_fs_close(cpu_state_T* state) {
    file_descriptor_t fd = (file_descriptor_t)state->rdi; // arg1

    file_descriptor_free(scheduler_get_current_process()->fd_array, fd);

    state->rax = STATUS_SUCCESS;
}

void syscall_handle_nx_fs_read(cpu_state_T* state) {
    file_descriptor_t fd     = (file_descriptor_t)state->rdi; // arg1
    uint64_t          offset = state->rsi;                    // arg2
    uint8_t*          mem    = (uint8_t*)state->rdx;          // arg3
    uint64_t          n      = state->rcx;                    // arg4

    process_T* process    = scheduler_get_current_process();
    uint64_t   read_bytes = 0;

    switch (fd) {
        case FILE_DESCRIPTOR_INVALID:
        case FILE_DESCRIPTOR_STDOUT:
        case FILE_DESCRIPTOR_STDERR: {
            break;
        }

        case FILE_DESCRIPTOR_STDIN: {
            read_bytes = pipe_read(&process->stdin, mem, n);
            break;
        }

        default: {
            vfs_node_T* node = file_descriptor_resolve(scheduler_get_current_process()->fd_array, fd);

            read_bytes = vfs_file_read(node, offset, n, mem);
            break;
        }
    }

    state->rax = read_bytes;
}

void syscall_handle_nx_fs_write(cpu_state_T* state) {
    file_descriptor_t fd     = (file_descriptor_t)state->rdi; // arg1
    uint64_t          offset = state->rsi;                    // arg2
    uint8_t*          mem    = (uint8_t*)state->rdx;          // arg3
    uint64_t          n      = state->rcx;                    // arg4

    process_T* process       = scheduler_get_current_process();
    uint64_t   written_bytes = 0;

    switch (fd) {
        case FILE_DESCRIPTOR_INVALID:
        case FILE_DESCRIPTOR_STDIN: {
            break;
        }

        case FILE_DESCRIPTOR_STDERR: {
            if (process->stderr != NULL) {
                written_bytes = pipe_write(process->stderr, mem, n);
            }
            break;
        }

        case FILE_DESCRIPTOR_STDOUT: {
            if (process->stdout != NULL) {
                written_bytes = pipe_write(process->stdout, mem, n);
            }
            break;
        }

        default: {
            vfs_node_T* node = file_descriptor_resolve(scheduler_get_current_process()->fd_array, fd);

            written_bytes = vfs_file_write(node, offset, n, mem);
            break;
        }
    }

    state->rax = written_bytes;
}

void syscall_handle_nx_fs_delete(cpu_state_T* state) {
    string_t path = (string_t)state->rdi; // arg1
    string_t len  = (string_t)state->rsi; // arg2

    vfs_node_T* node = vfs_resolve_path(&g_root_fs, path);
    if (node == NULL) {
        state->rax = STATUS_RESOURCE_NOT_AVAILABLE;
        return;
    }

    vfs_node_delete(node);

    state->rax = STATUS_SUCCESS;
}

void syscall_handle_nx_fs_list(cpu_state_T* state) {
    string_t  path       = (string_t)state->rdi;  // arg1
    uint64_t  len        = state->rsi;            // arg2
    uint8_t*  mem        = (uint8_t*)state->rdx;  // arg3
    uint32_t* needed_mem = (uint32_t*)state->rcx; // arg4

    vfs_node_T* node = vfs_resolve_path(&g_root_fs, path);
    if (node == NULL) {
        state->rax = STATUS_RESOURCE_NOT_AVAILABLE;
        return;
    }

    uint32_t position = 0;

    vfs_node_T* child = node->childs;
    while (child != NULL) {
        uint32_t strlen = string_length(child->name);

        if (mem != NULL) {
            memory_copy(child->name, &mem[position], strlen);
            mem[position + strlen] = '\0';
        }

        position += strlen + 1;
        child = child->next;
    }

    if (mem == NULL) {
        *needed_mem = position;
    }

    state->rax = STATUS_SUCCESS;
}


void syscall_handle_nx_mem_alloc(cpu_state_T* state) {
    uint8_t* addr  = (uint8_t*)state->rdi;  // arg1
    uint64_t n     = state->rsi;            // arg2
    uint64_t flags = (uint64_t)state->rdx;  // arg3

    if ((uint64_t)&addr[n * PFRAME_SIZE] > MEM_REGION_KERNEL ||
        (uint64_t)addr < MEM_REGION_PROCESS_USABLE) {
        state->rax = STATUS_PERMISSION_DENIED;
        return;
    }

    uint64_t map_flags = 0;
    if (flags & (1 << SYSCALL_NX_MMAP_FLAG_WRITE))   map_flags |= PM_FLAG_READ_WRITE;
    if (flags & (1 << SYSCALL_NX_MMAP_FLAG_NO_EXEC)) map_flags |= PM_FLAG_NO_EXECUTE;

    page_map_T* page_map = scheduler_get_current_process()->page_map;

    for (int i = 0; i < n; i++) {
        // check if the page is already mapped
        if (page_map_check_memory(page_map, &addr[i * PFRAME_SIZE])) {
            // unmap all the already mapped pages
            for (int j = 0; j < i; j++) {
                pframe_free(page_map_get_physical_address(page_map, &addr[j * PFRAME_SIZE]));
                page_map_unmap_memory(page_map, &addr[j * PFRAME_SIZE]);
            }
            state->rax = STATUS_RESOURCE_NOT_AVAILABLE;
            return;
        }
        page_map_map_memory(page_map, &addr[i * PFRAME_SIZE], pframe_request(), map_flags);
    }

    state->rax = STATUS_SUCCESS;
}

void syscall_handle_nx_mem_free(cpu_state_T* state) {
    void*    addr  = (void*)state->rdi;     // arg1
    uint64_t n     = state->rsi;            // arg2

    if ((uint64_t)&addr[n * PFRAME_SIZE] > MEM_REGION_KERNEL ||
        (uint64_t)addr < MEM_REGION_PROCESS_USABLE) {
        state->rax = STATUS_PERMISSION_DENIED;
        return;
    }

    page_map_T* page_map = scheduler_get_current_process()->page_map;

    for (int i = 0; i < n; i++) {
        pframe_free(page_map_get_physical_address(page_map, &addr[i * PFRAME_SIZE]));
        page_map_unmap_memory(page_map, &addr[i * PFRAME_SIZE]);
    }

    state->rax = STATUS_SUCCESS;
}


void syscall_handle_nx_proc_create(cpu_state_T* state) {
    process_config_T* conf    = (process_config_T*)state->rdi;  // arg1
    pid_t*            pid_ptr = (pid_t*)state->rsi;             // arg2

    process_T* parent = scheduler_get_current_process();

    vfs_node_T* node = file_descriptor_resolve(parent->fd_array, conf->executable);
    if (node == NULL || node->type != VFS_NODE_FILE) {
        state->rax = STATUS_RESOURCE_NOT_AVAILABLE;
        return;
    }

    uint8_t* buffer = memory_allocate(node->size);
    vfs_file_read(node, 0, node->size, buffer);

    elf_executable_T* exec = elf_executable_create(buffer);
    if (exec == NULL) {
        memory_free(buffer);
        state->rax = STATUS_NOT_SUPPORTED;
        return;
    }

    *pid_ptr = process_spawn(parent->id, conf->name, exec, buffer);
    if (*pid_ptr == PROCESS_NONE) {
        state->rax = STATUS_GENERIC_ERROR;
        return;
    }
    void* entry = (void*)(symbol_resolve_from_name(&exec->symbol_table, "_start")->address + MEM_REGION_PROCESS_EXEC);

    thread_spawn(*pid_ptr, entry);

    state->rax = STATUS_SUCCESS;
}

void syscall_handle_nx_proc_signal_send(cpu_state_T* state) {
    pid_t    pid    = (pid_t)state->rdi; // arg1
    uint64_t signal = state->rsi;        // arg2

    if (pid == PROCESS_NONE) {
        state->rax = STATUS_RESOURCE_NOT_AVAILABLE;
        return;
    }

    process_T* process = scheduler_get_process(pid);
    if (!process_is_child_of(process, scheduler_get_current_process())) {
        state->rax = STATUS_PERMISSION_DENIED;
        return;
    }

    process_signal(scheduler_get_process(pid), signal);
    state->rax = STATUS_SUCCESS;
}


void syscall_handle_nx_drv_register(cpu_state_T* state) {
    file_descriptor_t fd     = (file_descriptor_t)state->rdi; // arg1
    uint64_t*         id_ptr = (uint64_t*)state->rsi;         // arg2

    process_T*  process = scheduler_get_current_process();
    vfs_node_T* node    = file_descriptor_resolve(process->fd_array, fd);
    if (node == NULL) {
        log(LOG_ERROR, "failed to register driver (file descriptor error)");
        state->rax = STATUS_RESOURCE_NOT_AVAILABLE;
        return;
    }

    uint8_t* buffer = memory_allocate(node->size);
    vfs_file_read(node, 0, node->size, buffer);

    elf_executable_T* exec = elf_executable_create(buffer);
    if (exec == NULL) {
        memory_free(buffer);
        log(LOG_ERROR, "failed to register driver (elf parsing error)");
        state->rax = STATUS_NOT_SUPPORTED;
        return;
    }

    driver_T* drv = driver_register(exec, buffer);
    if (drv == NULL) {
        log(LOG_ERROR, "failed to register driver (registration error)");
        state->rax = STATUS_GENERIC_ERROR;
        return;
    }

    void (*drv_init)() = drv->base + symbol_resolve_from_name(&exec->symbol_table, "_init")->address;;
    drv_init();

    *id_ptr = drv->id;
    state->rax = STATUS_SUCCESS;
}

cpu_state_T* syscall_handle(cpu_state_T* state) {
    cpu_state_T*    return_state = state;
    syscall_group_E group_id     = (state->rax & 0xFF00) >> 8;
    uint64_t        arg1         = state->rdi;

    switch (group_id) {
        case SYSCALLS_FILES: {

            switch (state->rax) {
                case SYSCALL_NX_FS_OPEN: {
                    syscall_handle_nx_fs_open(state);
                    break;
                }
                case SYSCALL_NX_FS_CLOSE: {
                    syscall_handle_nx_fs_close(state);
                    break;
                }
                case SYSCALL_NX_FS_READ: {
                    syscall_handle_nx_fs_read(state);
                    break;
                }
                case SYSCALL_NX_FS_WRITE: {
                    syscall_handle_nx_fs_write(state);
                    break;
                }
                case SYSCALL_NX_FS_DELETE: {
                    syscall_handle_nx_fs_delete(state);
                    break;
                }
                case SYSCALL_NX_FS_LIST: {
                    syscall_handle_nx_fs_list(state);
                    break;
                }
            }
            break;
        }

        case SYSCALLS_MEMORY: {
            switch (state->rax) {
                case SYSCALL_NX_MEM_ALLOC: {
                    syscall_handle_nx_mem_alloc(state);
                    break;
                }
                case SYSCALL_NX_MEM_FREE: {
                    syscall_handle_nx_mem_free(state);
                    break;
                }
            }
            break;
        }

        case SYSCALLS_PROC: {
            switch (state->rax) {
                case SYSCALL_NX_PROC_CREATE: {
                    syscall_handle_nx_proc_create(state);
                    break;
                }
                case SYSCALL_NX_PROC_SIGNAL_SEND: {
                    syscall_handle_nx_proc_signal_send(state);
                    break;
                }
            }
            break;
        }

        case SYSCALLS_DRIVERS: {
            switch (state->rax) {
                case SYSCALL_NX_DRV_REGISTER: {
                    syscall_handle_nx_drv_register(state);
                    break;
                }
            }
            break;
        }

        case SYSCALLS_KERNEL: {

            // check if the current Process is permitted to perform a kernel syscall
            if (scheduler_is_initialized() &&
                scheduler_get_current_process()->id != PROCESS_KERNEL)
            {
                log(LOG_WARNING, "non kernel process[%d:%d] tried to perform a kernel syscall (permission denied)", scheduler_get_current_process(), scheduler_get_current_thread());
                state->rax = STATUS_PERMISSION_DENIED;
                return return_state;
            }

            switch (state->rax) {
                case SYSCALL_NX_KERNEL_SCHEDULER_START: {
                    return_state = scheduler_start(state);
                    break;
                }

                case SYSCALL_NX_KERNEL_PANIC: {
                    panic(state, (string_t)arg1);
                    break;
                }
            }
            break;
        }

        default: {
            log(LOG_WARNING, "unknown syscall group: 0x%xb", group_id);
            break;
        }
    }

    return return_state;
}