// This file is part of noxos and licensed under the MIT open source license

#include "proc/pipe.h"
#include "proc/process.h"

void pipe_init(pipe_T* pipe, process_T* receiver, void (*on_write)()) {
    pipe->stream         = stream_alloc(PIPE_STD_STREAM_SIZE);
    pipe->receiver       = receiver;
    pipe->senders_bitmap = bitmap_init(PIPE_MAX_SENDERS);
    pipe->on_write       = on_write;
}

void pipe_destruct(pipe_T* pipe) {
    bitmap_destruct(&pipe->senders_bitmap);
    stream_destruct(pipe->stream);
}


bool pipe_add_sender(pipe_T* pipe, process_T* sender) {
    for (int i = 0; i < PIPE_MAX_SENDERS; i++) {
        if (bitmap_get(&pipe->senders_bitmap, i)) continue;

        pipe->senders[i] = sender;
        bitmap_set(&pipe->senders_bitmap, i, true);
        return true;
    }

    return false;
}

void pipe_remove_sender(pipe_T* pipe, process_T* sender) {
    for (int i = 0; i < PIPE_MAX_SENDERS; i++) {
        if (!bitmap_get(&pipe->senders_bitmap, i)) continue;
        if (pipe->senders[i] != sender) continue;

        pipe->senders[i] = NULL;
        bitmap_set(&pipe->senders_bitmap, i, false);
    }
}

uint64_t pipe_write(pipe_T* pipe, void* buffer_in, uint64_t n) {
    if (pipe == NULL) return 0;
    uint64_t written_bytes = stream_write(pipe->stream, buffer_in, n);

    if (pipe->receiver != NULL && bitmap_get(&pipe->receiver->waiting, PROCESS_BLOCK_WAIT_PIPE_IN)) {
        bitmap_set(&pipe->receiver->waiting, PROCESS_BLOCK_WAIT_PIPE_IN, false);
        process_start(pipe->receiver);
    }

    if (written_bytes < n) {
        for (int i = 0; i < PIPE_MAX_SENDERS; i++) {
            if (!bitmap_get(&pipe->senders_bitmap, i)) continue;

            bitmap_set(&pipe->senders[i]->waiting, PROCESS_BLOCK_WAIT_PIPE_OUT, true);
            process_pause(pipe->senders[i]);
        }
    }

    if (pipe->on_write != NULL) {
        pipe->on_write();
    }

    return written_bytes;
}

uint64_t pipe_read(pipe_T* pipe, void* buffer_out, uint64_t n) {
    if (pipe == NULL) return 0;

    uint64_t read_bytes = stream_read(pipe->stream, buffer_out, n);

    for (int i = 0; i < PIPE_MAX_SENDERS; i++) {
        if (!bitmap_get(&pipe->senders_bitmap, i)) continue;
        if (!bitmap_get(&pipe->senders[i]->waiting, PROCESS_BLOCK_WAIT_PIPE_OUT)) break;

        bitmap_set(&pipe->senders[i]->waiting, PROCESS_BLOCK_WAIT_PIPE_OUT, false);
        process_start(pipe->senders[i]);
    }

    if (pipe->receiver != NULL && read_bytes < n) {
        bitmap_set(&pipe->receiver->waiting, PROCESS_BLOCK_WAIT_PIPE_IN, true);
        process_pause(pipe->receiver);
    }

    return read_bytes;
}
