// This file is part of noxos and licensed under the MIT open source license

#include "proc/process.h"
#include "proc/scheduler.h"
#include "utils/memory.h"
#include "utils/math.h"
#include "utils/logger.h"
#include "mm/region.h"

void process_kernel_spawn(elf_executable_T* executable) {
    process_T* process = memory_allocate(sizeof(process_T));
    process->num_threads = 0;
    process->threads     = NULL;
    process->parent      = NULL;
    process->executable  = executable;
    process->fd_array    = file_descriptor_array_alloc();
    process->stdout      = NULL;
    process->stderr      = NULL;
    process->waiting     = bitmap_init(PROCESS_BLOCK_ENUM_SIZE);

    pipe_init(&process->stdin, process, NULL);
    memory_copy("kernel", process->name, 7);

    process->page_map = g_kernel_page_map;

    // the processes' id, chunk_id and chunk fields will be set by the scheduler

    scheduler_register_process(process);
}

pid_t process_spawn(pid_t parent, string_t name, elf_executable_T* executable, void* buffer) {
    process_T* process = memory_allocate(sizeof(process_T));

    process->num_threads = 0;
    process->threads     = NULL;
    process->thread_ids  = bitmap_init(MAX_THREADS_PER_PROCESS);
    process->parent      = scheduler_get_process(parent);
    process->executable  = executable;
    process->fd_array    = file_descriptor_array_alloc();
    process->stdout      = NULL;
    process->stderr      = NULL;
    process->waiting     = bitmap_init(PROCESS_BLOCK_ENUM_SIZE);

    pipe_init(&process->stdin, process, NULL);
    memory_copy(name, process->name, MIN(string_length(name)+1, 127));

    process->page_map = page_map_create();
    elf_mappings_apply(executable->mappings, executable->num_mappings, buffer, MEM_REGION_PROCESS_EXEC, process->page_map);

    // the processes' id, chunk_id and chunk fields will be set by the scheduler

    return scheduler_register_process(process);
}

void process_signal(process_T* process, process_signals_E signal) {
    switch (signal) {
        case PROCESS_SIGNAL_START: {
            process_start(process);
            break;
        }
        case PROCESS_SIGNAL_PAUSE: {
            process_pause(process);
            break;
        }
        case PROCESS_SIGNAL_KILL: {
            process_kill(process);
            break;
        }
        default: break;
    }
}

bool process_is_child_of(process_T* process, process_T* parent) {
    process_T* tmp = process->parent;
    while (tmp != NULL) {
        if (tmp == parent) return true;
        tmp = tmp->parent;
    }
    return false;
}

int32_t process_get_thread_id(process_T* process) {
    for (int i = 0; i < MAX_THREADS_PER_PROCESS; i++) {
        if (!bitmap_get(&process->thread_ids, i)) {
            bitmap_set(&process->thread_ids, i, true);
            return i;
        }
    }
    return THREAD_ID_INVALID;
}

void process_clear_thread_id(process_T* process, uint32_t id) {
    bitmap_set(&process->thread_ids, id, false);
}

void process_pause_pid(pid_t pid) {
    process_pause(scheduler_get_process(pid));
}

void process_pause(process_T* process) {
    scheduler_pause_process(process);
}

void process_start_pid(pid_t pid) {
    process_start(scheduler_get_process(pid));
}

void process_start(process_T* process) {
    scheduler_start_process(process);
}

void process_kill_pid(pid_t pid) {
    process_kill(scheduler_get_process(pid));
}

void process_kill(process_T* process) {
    scheduler_kill_process(process);

    pipe_destruct(&process->stdin);
    file_descriptor_array_destruct(process->fd_array);
    elf_executable_destruct(process->executable);
    page_map_destruct(process->page_map);
    bitmap_destruct(&process->thread_ids);

    memory_free(process);
}