// This file is part of noxos and licensed under the MIT open source license

#include "proc/scheduler.h"
#include "utils/memory.h"
#include "utils/logger.h"
#include "utils/core.h"
#include "drivers/builtin/graphics/renderer.h"
#include "platform/syscall.h"

scheduler_T         g_scheduler;
graphics_buffer_T*  g_scheduler_info_buffer;

void scheduler_update_info() {
    uint64_t size = format_size("Processes[%d]\n  Threads[%d]", g_scheduler.num_processes, g_scheduler.num_threads);
    char buffer[size];
    format(buffer, "Processes[%d]\n  Threads[%d]", g_scheduler.num_processes, g_scheduler.num_threads);

    graphics_buffer_draw_string(g_scheduler_info_buffer, 0, 0, g_color_palette[COLOR_PAL_GREEN], buffer);
}

void scheduler_dump_info(process_T* process, uint8_t indent) {
    char buf[indent+1];
    memory_set(buf, ' ', indent);
    buf[indent] = '\0';

    log(LOG_NONE, "%s<%d>  %s  [%d]", buf, process->id, process->name, process->num_threads);

    process_T* child = process->childs;
    while (child != NULL) {
        scheduler_dump_info(child, indent+2);

        child = child->next;
    }
}

scheduler_processes_chunk_T* scheduler_processes_chunk_create(scheduler_processes_chunk_T* prev) {
    scheduler_processes_chunk_T* chunk = memory_allocate(sizeof(scheduler_processes_chunk_T));

    chunk->processes        = memory_allocate(SCHEDULER_PROCESS_CHUNK_SIZE * sizeof(process_T*));
    chunk->prev             = prev;
    chunk->next             = NULL;
    chunk->processes_bitmap = bitmap_init(SCHEDULER_PROCESS_CHUNK_SIZE);
    chunk->num_free_pids    = SCHEDULER_PROCESS_CHUNK_SIZE;

    return chunk;
}

void scheduler_init(boot_info_T* boot_info) {
    g_scheduler.num_threads     = 0;
    g_scheduler.num_processes   = 0;
    g_scheduler.running_thread  = NULL;
    g_scheduler.blocked         = false;

    g_scheduler.processes       = scheduler_processes_chunk_create(NULL);

    g_scheduler_info_buffer = graphics_buffer_request(graphics_renderer_get_width()-100, 0, 100, 200, GRAPHICS_BUFFER_STANDARD);

    process_kernel_spawn(elf_executable_create(boot_info->kernel_file->address));

    syscall_perform(SYSCALL_NX_KERNEL_SCHEDULER_START, 0, 0, 0, 0);
}

cpu_state_T* scheduler_start(cpu_state_T* state) {

    thread_T* thread = thread_spawn_from_state(PROCESS_KERNEL, state);
    thread->global_prev = thread;
    thread->global_next = thread;
    g_scheduler.running_thread = thread;


    scheduler_start_thread(thread);

    g_scheduler.initialized = true;

    return scheduler_switch_context(state);
}

bool scheduler_is_initialized() {
    return g_scheduler.initialized;
}

void scheduler_queue_add_thread(thread_T* thread) {
    if (g_scheduler.running_thread == NULL) { return; }

    thread->global_prev                                  = g_scheduler.running_thread;
    thread->global_next                                  = g_scheduler.running_thread->global_next;
    g_scheduler.running_thread->global_next->global_prev = thread;
    g_scheduler.running_thread->global_next              = thread;
}

void scheduler_queue_remove_thread(thread_T* thread) {
    if (thread->global_prev == NULL || thread->global_next == NULL) { return; }

    thread->global_prev->global_next = thread->global_next;
    thread->global_next->global_prev = thread->global_prev;
//    thread->global_prev              = NULL;
//    thread->global_next              = NULL;
}

thread_T* scheduler_register_thread(thread_T* thread) {
    thread->local_prev = NULL;
    thread->local_next = NULL;

    if (thread->process->threads != NULL) {
        thread->local_next             = thread->process->threads;
        thread->local_next->local_prev = thread;
    }

    thread->process->threads = thread;

    thread->process->num_threads++;
    g_scheduler.num_threads++;

    scheduler_update_info();

    return thread;
}

void scheduler_pause_thread(thread_T* thread) {
    scheduler_queue_remove_thread(thread);
    scheduler_update_info();
}

void scheduler_start_thread(thread_T* thread) {
    scheduler_queue_add_thread(thread);
    scheduler_update_info();
}

void scheduler_kill_thread(thread_T* thread) {
    scheduler_queue_remove_thread(thread);

    if (thread->process->threads == thread) {
        thread->process->threads = thread->local_next;
    }

    if (thread->local_next != NULL) {
        thread->local_next->local_prev = thread->local_prev;
    }

    if (thread->local_prev != NULL) {
        thread->local_prev->local_next = thread->local_next;
    }

    thread->process->num_threads--;
    g_scheduler.num_threads--;

    scheduler_update_info();
}

pid_t scheduler_register_process(process_T* process) {
    pid_t pid = 1;  // offset pit by one, so this won't return a pid that is PROCESS_NONE

    // find a process chunk, that has free slots
    scheduler_processes_chunk_T* chunk = g_scheduler.processes;
    while (chunk->num_free_pids == 0) {
        if (chunk->next == NULL) {
            chunk->next = scheduler_processes_chunk_create(chunk);
        }
        chunk  = chunk->next;
        pid   += SCHEDULER_PROCESS_CHUNK_SIZE;
    }

    for (uint32_t i = 0; i < SCHEDULER_PROCESS_CHUNK_SIZE; i++) {
        if (bitmap_get(&chunk->processes_bitmap, i)) {
            continue;
        }

        bitmap_set(&chunk->processes_bitmap, i, true);
        chunk->processes[i] = process;
        process->chunk      = chunk;
        process->chunk_id   = i;
        process->id         = pid + i;
        process->childs     = NULL;
        process->prev       = NULL;

        if (process->parent != NULL) {
            process->next = process->parent->childs;

            if (process->next != NULL) {
                process->next->prev = process;
            }

            process->parent->childs = process;
        } else {
            process->next = NULL;
        }

        chunk->num_free_pids--;
        g_scheduler.num_processes++;

        log(LOG_INFO, "<Scheduler> Registered Process %d", pid + i);
        scheduler_update_info();

        return pid + i;
    }

    // if this happens, something went really wrong
    log(LOG_ERROR, "<Scheduler> failed to register process (no free slots in chunk)");
    return PROCESS_NONE;
}

void scheduler_pause_process(process_T* process) {
    thread_T* thread = process->threads;
    while (thread != NULL) {
        thread_T* next = thread->local_next;
        thread_pause(thread);
        thread = next;
    }
}

void scheduler_start_process(process_T* process) {
    thread_T* thread = process->threads;
    while (thread != NULL) {
        thread_T* next = thread->local_next;
        thread_start(thread);
        thread = next;
    }
}

void scheduler_kill_process(process_T* process) {
    scheduler_processes_chunk_T* chunk = process->chunk;

    bitmap_set(&chunk->processes_bitmap, process->chunk_id, false);
    chunk->processes[process->chunk_id] = NULL;

    process_T* child = process->childs;
    while (child != NULL) {
        process_T* next = child->next;
        process_kill(child);
        child = next;
    }

    thread_T* thread = process->threads;
    while (thread != NULL) {
        thread_T* next = thread->local_next;
        thread_kill(thread);
        thread = next;
    }

    if (process->parent != NULL) {
        if (process->parent->childs == process) {
            process->parent->childs = process->next;
        }

        if (process->next != NULL) {
            process->next->prev = process->prev;
        }

        if (process->prev != NULL) {
            process->prev->next = process->next;
        }
    }

    chunk->num_free_pids++;
    g_scheduler.num_processes--;

    log(LOG_INFO, "<Scheduler> Killed Process %d", process->id);
    scheduler_update_info();
}

process_T* scheduler_get_process(pid_t pid) {
    if (pid == PROCESS_NONE) {
        return NULL;
    }

    pid--;  // apply pid offset to make PROCESS_NONE work

    uint32_t num_chunks     = pid / SCHEDULER_PROCESS_CHUNK_SIZE;
    uint32_t index_in_chunk = pid % SCHEDULER_PROCESS_CHUNK_SIZE;

    scheduler_processes_chunk_T* chunk = g_scheduler.processes;
    if (chunk == NULL) {
        return NULL;
    }
    for (uint32_t i = 0; i < num_chunks; i++) {
        chunk = chunk->next;

        if (chunk == NULL) {
            return NULL;
        }
    }

    if (!bitmap_get(&chunk->processes_bitmap, index_in_chunk)) {
        return NULL;
    }

    return chunk->processes[index_in_chunk];
}

thread_T* scheduler_get_current_thread() {
    if (!g_scheduler.initialized) {
        return NULL;
    }

    return g_scheduler.running_thread;
}

process_T* scheduler_get_current_process() {
    thread_T* thread = scheduler_get_current_thread();
    if (thread == NULL) {
        return NULL;
    }

    return thread->process;
}

cpu_state_T* scheduler_switch_context(cpu_state_T* state) {
    if (!g_scheduler.initialized) {
        return state;
    }

    CORE_HALT_WHILE(g_scheduler.blocked)
    g_scheduler.blocked = true;

    thread_T* old_thread = g_scheduler.running_thread;
    thread_T* new_thread = g_scheduler.running_thread->global_next;

    if (old_thread->cpu_time > 0) {
        old_thread->state = state;
    }

    // dirty map 0x0000010000000000 <--> 0x0000020000000000
    for (int i = 2; i < 5; i++) {
        g_kernel_page_map->entries[i] = new_thread->process->page_map->entries[i];
    }

    new_thread->cpu_time += 1;
    g_scheduler.running_thread = new_thread;

    g_scheduler.blocked = false;

    return new_thread->state;
}