// This file is part of noxos and licensed under the MIT open source license

#include "proc/thread.h"
#include "proc/scheduler.h"
#include "mm/page_map.h"
#include "mm/page_frame.h"
#include "mm/region.h"
#include "platform/gdt.h"
#include "utils/memory.h"
#include "utils/logger.h"

void thread_end_point() {
    // this has to kill the thread in future
    while (1) asm("hlt");
}

thread_T* thread_spawn(pid_t process, void* function) {
    thread_T* thread = memory_allocate(sizeof(thread_T));

    thread->process = scheduler_get_process(process);
    if (thread->process == NULL) {
        memory_free(thread);
        log(LOG_WARNING, "<Thread> failed to spawn thread for process %d (invalid pid)", process);
        return NULL;
    }

    int32_t id = process_get_thread_id(thread->process);
    if (id == THREAD_ID_INVALID) {
        memory_free(thread);
        log(LOG_WARNING, "<Thread> failed to spawn thread for process %d (no free thread id)", process);
        return NULL;
    }
    thread->local_id = (uint32_t)id;

    uint8_t* base = (uint8_t*)((process == PROCESS_KERNEL ? MEM_REGION_KERNEL_THREAD_BASE : MEM_REGION_PROCESS_THREAD_BASE) + (MEM_REGION_THREAD_OFFSET * thread->local_id));

    thread->cpu_time   = 0;

    thread->stack_size = PFRAME_SIZE * 4;
    thread->stack      = base;
    for (int i = 0; i < 4; i++) {
        void* physical_address = pframe_request();
        page_map_map_memory(thread->process->page_map, &base[i * PFRAME_SIZE], physical_address, PM_FLAG_READ_WRITE);
        page_map_map_memory(g_kernel_page_map, MEM_REGION_KERNEL_STACK_DUMMY + i * PFRAME_SIZE, physical_address, PM_FLAG_READ_WRITE);
    }

    *((uint64_t*)(MEM_REGION_KERNEL_STACK_DUMMY + (thread->stack_size - sizeof(cpu_state_T) - 8))) = (uint64_t)thread_end_point;

    cpu_state_T state = {
            .cr3 = (uint64_t)thread->process->page_map,
            .rax = 0,
            .rbx = 0,
            .rcx = 0,
            .rdx = 0,
            .rsi = 0,
            .rdi = 0,
            .rbp = (uint64_t)thread->stack + (thread->stack_size - sizeof(cpu_state_T) - 8),
            .interrupt_id = 0,
            .error_code = 0,
            .rip = (uint64_t)function,
            .cs = GDT_SELECTOR_KERNEL_CODE,
            .flags = 1 << CPU_FLAG_INTERRUPT_ENABLE,
            .rsp = (uint64_t)thread->stack + (thread->stack_size - sizeof(cpu_state_T) - 8),
            .ss = GDT_SELECTOR_KERNEL_DATA
    };

    memory_copy(&state, MEM_REGION_KERNEL_STACK_DUMMY + (thread->stack_size - sizeof(cpu_state_T)), sizeof(cpu_state_T));

    thread->state = thread->stack + (thread->stack_size - sizeof(cpu_state_T));

    for (int i = 0; i < 4; i++) {
        page_map_unmap_memory(g_kernel_page_map, MEM_REGION_KERNEL_STACK_DUMMY + i * PFRAME_SIZE);
    }

    return scheduler_register_thread(thread);
}

thread_T* thread_spawn_from_state(pid_t process, cpu_state_T* state) {
    thread_T* thread = memory_allocate(sizeof(thread_T));

    thread->process = scheduler_get_process(process);
    if (thread->process == NULL) {
        memory_free(thread);
        log(LOG_WARNING, "<Thread> failed to spawn thread for process %d (invalid pid)", process);
        return NULL;
    }

    thread->stack_size = 0;
    thread->stack      = NULL;
    thread->cpu_time   = 0;

    thread->state = state;

    return scheduler_register_thread(thread);
}

void thread_start(thread_T* thread) {
    scheduler_start_thread(thread);
}

void thread_pause(thread_T* thread) {
    scheduler_pause_thread(thread);
}

void thread_kill(thread_T* thread) {
    scheduler_kill_thread(thread);

    process_clear_thread_id(thread->process, thread->local_id);

    uint8_t* base = (uint8_t*)((thread->process->id == PROCESS_KERNEL ? MEM_REGION_KERNEL_THREAD_BASE : MEM_REGION_PROCESS_THREAD_BASE) + (MEM_REGION_THREAD_OFFSET * thread->local_id));

    if (thread->stack != NULL) {
        for (int i = 0; i < 4; i++) {
            void* pframe = page_map_get_physical_address(thread->process->page_map, &base[PFRAME_SIZE * i]);
            pframe_free(pframe);
            page_map_unmap_memory(thread->process->page_map, &base[PFRAME_SIZE * i]);
        }
    }

    memory_free(thread);
}