// This file is part of noxos and licensed under the MIT open source license

#include "utils/memory.h"
#include "utils/logger.h"
#include "mm/heap.h"

heap_T g_kernel_heap;

void memory_copy(void* source, void* destination, uint32_t num) {
    uint32_t  num_64 = num / 8;
    uint64_t* src_64 = (uint64_t*)source;
    uint64_t* dst_64 = (uint64_t*)destination;

    for (uint32_t i = 0; i < num_64; i++) {
        dst_64[i] = src_64[i];
    }

    uint32_t num_8 = num % 8;
    uint8_t* src_8 = &((uint8_t*)source)[num - num_8];
    uint8_t* dst_8 = &((uint8_t*)destination)[num - num_8];

    for (uint32_t i = 0; i < num_8; i++) {
        dst_8[i] = src_8[i];
    }
}

void memory_set(void* destination, uint8_t data, uint32_t num) {
    uint32_t  num_64  = num / 8;
    uint64_t* dst_64  = (uint64_t*)destination;
    uint64_t  data_64 = data;

    data_64 |= (data_64 << 8) | (data_64 << 16) | (data_64 << 24) | (data_64 << 32) | (data_64 << 40) | (data_64 << 48) | (data_64 << 56);

    for (uint32_t i = 0; i < num_64; i++) {
        dst_64[i] = data_64;
    }

    uint32_t num_8 = num % 8;
    uint8_t* dst_8 = &((uint8_t*)destination)[num - num_8];

    for (uint32_t i = 0; i < num_8; i++) {
        dst_8[i] = data;
    }
}

bool memory_compare(void* a, void* b, uint32_t num) {
    uint32_t  num_64 = num / 8;
    uint64_t* a_64   = (uint64_t*)a;
    uint64_t* b_64   = (uint64_t*)b;

    for (uint32_t i = 0; i < num_64; i++) {
        if (a_64[i] != b_64[i]) {
            return false;
        }
    }

    uint32_t num_8 = num % 8;
    uint8_t* a_8 = &((uint8_t*)a)[num - num_8];
    uint8_t* b_8 = &((uint8_t*)b)[num - num_8];

    for (uint32_t i = 0; i < num_8; i++) {
        if (a_8[i] != b_8[i]) {
            return false;
        }
    }

    return true;
}

void* memory_allocate(uint64_t size) {
    return heap_memory_allocate(&g_kernel_heap, size);
}

void memory_free(void* address) {
    return heap_memory_free(&g_kernel_heap, address);
}

void memory_allocator_init(void* base) {
    if (g_kernel_heap.start != NULL) { return; }

    heap_init(&g_kernel_heap, base);
}

void memory_hexdump(uint8_t* address, uint64_t num) {
    uint64_t rows      = num / 8;
    uint8_t  remainder = num % 8;

    for (uint64_t i = 0; i < rows; i++) {
        char buffer[10] = ".... ....";
        for (uint8_t j = 0; j < 8; j++) {
            if (string_is_char_text((char)address[j])) {
                buffer[(j < 4) ? j : (j + 1)] = (char)address[j];
            }
        }

        log(LOG_NONE, "<Hexdump>  0x%x -> %xb %xb %xb %xb  %xb %xb %xb %xb  [%s]",
            address, address[0], address[1], address[2], address[3], address[4], address[5], address[6], address[7], buffer);

        address += 8;
    }


    char buffer[10] = "         ";
    for (uint8_t j = 0; j < remainder; j++) {
        if (string_is_char_text((char)address[j])) {
            buffer[(j < 4) ? j : (j + 1)] = (char)address[j];
        } else {
            buffer[(j < 4) ? j : (j + 1)] = '.';
        }
    }

    // this is not a beautiful solution, but it should be easy to understand / maintain
    switch (remainder) {
        case 0: {
            break;
        }
        case 1: {
            log(LOG_NONE, "<Hexdump>  0x%x -> %xb                        [%s]",
                address, address[0], buffer);
            break;
        }
        case 2: {
            log(LOG_NONE, "<Hexdump>  0x%x -> %xb %xb                     [%s]",
                address, address[0], address[1], buffer);
            break;
        }
        case 3: {
            log(LOG_NONE, "<Hexdump>  0x%x -> %xb %xb %xb                  [%s]",
                address, address[0], address[1], address[2], buffer);
            break;
        }
        case 4: {
            log(LOG_NONE, "<Hexdump>  0x%x -> %xb %xb %xb %xb               [%s]",
                address, address[0], address[1], address[2], address[3], buffer);
            break;
        }
        case 5: {
            log(LOG_NONE, "<Hexdump>  0x%x -> %xb %xb %xb %xb  %xb           [%s]",
                address, address[0], address[1], address[2], address[3], address[4], buffer);
            break;
        }
        case 6: {
            log(LOG_NONE, "<Hexdump>  0x%x -> %xb %xb %xb %xb  %xb %xb        [%s]",
                address, address[0], address[1], address[2], address[3], address[4], address[5], buffer);
            break;
        }
        case 7: {
            log(LOG_NONE, "<Hexdump>  0x%x -> %xb %xb %xb %xb  %xb %xb %xb     [%s]",
                address, address[0], address[1], address[2], address[3], address[4], address[5], address[6], buffer);
            break;
        }
        default: {
            break;
        }
    }
}