// This file is part of noxos and licensed under the MIT open source license

#include "utils/panic.h"
#include "utils/logger.h"
#include "utils/core.h"
#include "platform/exceptions.h"
#include "mm/stack.h"
#include "mm/region.h"
#include "proc/scheduler.h"

void panic_log_paging_info(cpu_state_T* state) {
    log(LOG_NONE, "Paging Info:");
    log(LOG_NONE, "    Page Map: 0x%x", state->cr3);

    if (state->interrupt_id == EXCEPTION_PAGE_FAULT) {
        uint64_t cr2;
        asm("mov %%cr2, %0" : "=r" (cr2));

        log(LOG_NONE, "    Page Fault Info:");
        log(LOG_NONE, "        %s", state->error_code & (1 << 0) ? "Page protection violation" : "Non present page access");
        log(LOG_NONE, "        %s", state->error_code & (1 << 1) ? "Write" : "Read");
        log(LOG_NONE, "        Address: 0x%x", cr2);
    }

    log(LOG_NONE, "");  // newline
}

void panic_log_eflags(cpu_state_T* state) {
    if (state->flags) {
        log(LOG_NONE, "CPU Flags:");
        if (state->flags & (1 << CPU_FLAG_CARRY))                     { log(LOG_NONE, "    Carry"); }
        if (state->flags & (1 << CPU_FLAG_PARITY))                    { log(LOG_NONE, "    Parity"); }
        if (state->flags & (1 << CPU_FLAG_AUXILIARY))                 { log(LOG_NONE, "    Auxiliary"); }
        if (state->flags & (1 << CPU_FLAG_ZERO))                      { log(LOG_NONE, "    Zero"); }
        if (state->flags & (1 << CPU_FLAG_SIGN))                      { log(LOG_NONE, "    Sign"); }
        if (state->flags & (1 << CPU_FLAG_TRAP))                      { log(LOG_NONE, "    Trap"); }
        if (state->flags & (1 << CPU_FLAG_INTERRUPT_ENABLE))          { log(LOG_NONE, "    Interrupt Enable"); }
        if (state->flags & (1 << CPU_FLAG_DIRECTION))                 { log(LOG_NONE, "    Direction"); }
        if (state->flags & (1 << CPU_FLAG_OVERFLOW))                  { log(LOG_NONE, "    Overflow"); }
        if (state->flags & (1 << CPU_FLAG_IO_PRIVILEGE_0))            { log(LOG_NONE, "    IO Privilege 0"); }
        if (state->flags & (1 << CPU_FLAG_IO_PRIVILEGE_1))            { log(LOG_NONE, "    IO Privilege 1"); }
        if (state->flags & (1 << CPU_FLAG_NESTED_TASK))               { log(LOG_NONE, "    Nested Task"); }
        if (state->flags & (1 << CPU_FLAG_RESUME))                    { log(LOG_NONE, "    Resume"); }
        if (state->flags & (1 << CPU_FLAG_VIRTUAL_8086))              { log(LOG_NONE, "    Virtual 8086"); }
        if (state->flags & (1 << CPU_FLAG_ALIGNMENT_CHECK))           { log(LOG_NONE, "    Alignment Check"); }
        if (state->flags & (1 << CPU_FLAG_VIRTUAL_INTERRUPT))         { log(LOG_NONE, "    Virtual Interrupt"); }
        if (state->flags & (1 << CPU_FLAG_VIRTUAL_INTERRUPT_PENDING)) { log(LOG_NONE, "    Virtual Interrupt Pending"); }
        if (state->flags & (1 << CPU_FLAG_CPUID))                     { log(LOG_NONE, "    CPUID"); }

        log(LOG_NONE, "");  // newline
    }
}

void panic_log_registers(cpu_state_T* state) {
    log(LOG_NONE, "CPU Registers:");
    log(LOG_NONE, "    RIP: 0x%x   RAX: 0x%x   RBX: 0x%x", state->rip, state->rax, state->rbx);
    log(LOG_NONE, "    RCX: 0x%x   RDX: 0x%x   RSI: 0x%x", state->rcx, state->rdx, state->rsi);
    log(LOG_NONE, "    RDI: 0x%x   RBP: 0x%x   RSP: 0x%x\n", state->rdi, state->rbp, state->rsp);
}

void panic_log_call_stack(cpu_state_T* state) {
    process_T* process = state->rip >= MEM_REGION_KERNEL ? scheduler_get_process(PROCESS_KERNEL) : scheduler_get_current_process();
    if (process == NULL || process->executable == NULL) {
        log(LOG_NONE, "Call Stack not resolvable (no symbol data access)");
        return;
    }

    log(LOG_NONE, "Call Stack:");
    stack_trace_call_stack(state->rbp);

    symbol_T* symbol = symbol_resolve_from_rip(&process->executable->symbol_table, state->rip);
    if (symbol == NULL) {
        log(LOG_NONE, "    0x%x       ->  <failed to resolve symbol>  (symbol NULL reference)", state->rip);
        return;
    }

    stack_dump_call_info(state->rip, symbol);
}

void panic(cpu_state_T* state, string_t message) {
    log(LOG_ERROR, "!=====[ KERNEL PANIC ]=====!");
    log(LOG_NONE, "Error Message: %s", message);

    if (state == NULL) {
        log(LOG_WARNING, "No detailed Information available (cpu_state null reference)");
        CORE_HALT_FOREVER
    }

    log(LOG_NONE, "Interrupt ID: 0x%xb", state->interrupt_id);
    log(LOG_NONE, "Error Code: 0b%.32b\n", state->error_code);


    panic_log_paging_info(state);
    panic_log_eflags(state);
    panic_log_registers(state);
    panic_log_call_stack(state);

    CORE_HALT_FOREVER
}