// This file is part of noxos and licensed under the MIT open source license

#include "utils/string.h"
#include "utils/memory.h"
#include "utils/math.h"

const char g_hex_chars[16] = "0123456789ABCDEF";

uint32_t string_length(string_t string) {
    return string_find_next(string, '\0');
}

bool string_compare(string_t a, string_t b) {
    uint32_t a_len = string_length(a);

    if (a_len != string_length(b)) {
        return false;
    }
    return memory_compare(a, b, a_len);
}

uint32_t string_find_next(string_t string, char chr) {
    uint32_t n = 0;

    while(*string != chr && *string != '\0') {
        string++;
        n++;
    }

    return n;
}

uint32_t string_find_last(string_t string, char chr) {
    uint32_t n = string_length(string);

    string += n;
    while(*string != chr && n != 0) {
        string--;
        n--;
    }

    return n;
}

void string_add_ext_command(string_t string, string_ext_command_E command, int num_data, uint8_t data[]) {
    string_ext_header_T* header = (string_ext_header_T*)string;
    header->shift_in = 0x0F;
    header->length   = num_data;
    header->command  = command;
    memory_copy(data, &header->data, num_data);
    *(&(&header->data)[num_data]) = 0x0E;
}

uint64_t variadic_format_size(string_t string, va_list args) {
    const char* buffer_in = (const char*)string;

    uint64_t length = 0;
    while(*buffer_in != '\0') {
        if (*buffer_in == '%') {
            uint8_t numeric_specifier = 0;
            buffer_in++;
            if (*buffer_in == '.') {
                buffer_in++;
                if (*buffer_in == '*') {
                    buffer_in++;
                    numeric_specifier = va_arg(args, uint64_t);
                } else {
                    while (*buffer_in > '0' && *buffer_in < '9') {
                        numeric_specifier *= 10;
                        numeric_specifier += *buffer_in - '0';
                        buffer_in++;
                    }
                }
            }

            switch(*buffer_in) {
                case '%': {
                    length++;
                    break;
                }
                case 's': {
                    // string
                    string_t str = va_arg(args, string_t);
                    if (str == NULL) {
                        str = "<string NULL reference>";
                    }

                    uint64_t strlen = string_length(str);
                    if (numeric_specifier > 0) {
                        strlen = MIN(strlen, numeric_specifier);
                    }

                    length += strlen;
                    break;
                }
                case 'c': {
                    // char
                    char tmp = (char)va_arg(args, int);
                    length++;
                    break;
                }
                case 'u': {
                    // decimal unsigned int
                    char str[32];
                    string_unsigned_dec_to_alpha(str, va_arg(args, uint64_t));
                    length += string_length(str);
                    break;
                }
                case 'd': {
                    // decimal signed int
                    char str[32];
                    string_dec_to_alpha(str, va_arg(args, int64_t));
                    length += string_length(str);
                    break;
                }
                case 'x': {
                    // Hexadecimal int
                    char str[32];

                    buffer_in++;
                    switch(*buffer_in) {
                        case 'b': {
                            // byte
                            string_hex_8bit_to_alpha(str, va_arg(args, int));
                            break;
                        }
                        case 'w': {
                            // word
                            string_hex_16bit_to_alpha(str, va_arg(args, int));
                            break;
                        }
                        case 'd': {
                            // dword
                            string_hex_32bit_to_alpha(str, va_arg(args, uint32_t));
                            break;
                        }
                        case 'q': {
                            // qword
                            string_hex_64bit_to_alpha(str, va_arg(args, uint64_t));
                            break;
                        }
                        default: {
                            string_hex_64bit_to_alpha(str, va_arg(args, uint64_t));
                            buffer_in--;
                            break;
                        }
                    }
                    length += string_length(str);
                    break;
                }
                case '?': {
                    char str[7];
                    string_bool_to_alpha(str, (bool)va_arg(args, int));
                    length += string_length(str);
                    break;
                }
                case 'b': {
                    if (numeric_specifier == 0) {
                        numeric_specifier = 64;
                    }

                    char str[65];
                    string_bin_to_alpha(str, numeric_specifier, va_arg(args, uint64_t));
                    length += string_length(str);
                    break;
                }
            }
        } else {
            length++;
        }
        buffer_in++;
    }
    length++;

    return length;
}

uint64_t format_size(string_t string, ...) {
    va_list args;
    va_start(args, string);

    uint64_t size = variadic_format_size(string, args);

    va_end(args);
    return size;
}

void variadic_format(string_t output, string_t string, va_list args) {
    char*       buffer_out =       (char*)output;
    const char* buffer_in  = (const char*)string;

    while(*buffer_in != '\0') {
        if (*buffer_in == '%') {
            uint8_t numeric_specifier = 0;
            buffer_in++;
            if (*buffer_in == '.') {
                buffer_in++;
                if (*buffer_in == '*') {
                    buffer_in++;
                    numeric_specifier = va_arg(args, uint64_t);
                } else {
                    while (*buffer_in > '0' && *buffer_in < '9') {
                        numeric_specifier *= 10;
                        numeric_specifier += *buffer_in - '0';
                        buffer_in++;
                    }
                }
            }

            switch(*buffer_in) {
                case '%': {
                    *buffer_out = '%';
                    buffer_out++;
                    break;
                }
                case 's': {
                    // string
                    string_t str = va_arg(args, string_t);
                    if (str == NULL) {
                        str = "<string NULL reference>";
                    }
                    uint64_t strlen = string_length(str);

                    if (numeric_specifier > 0) {
                        strlen = MIN(strlen, numeric_specifier);
                    }

                    memory_copy((void*)str, buffer_out, strlen);
                    buffer_out += strlen;
                    break;
                }
                case 'c': {
                    // char
                    *buffer_out = (char)va_arg(args, int);
                    buffer_out++;
                    break;
                }
                case 'u': {
                    // decimal unsigned int
                    char str[32];
                    string_unsigned_dec_to_alpha(str, va_arg(args, uint64_t));
                    uint64_t strlen = string_length(str);

                    memory_copy((void*)str, buffer_out, strlen);
                    buffer_out += strlen;
                    break;
                }
                case 'd': {
                    // decimal signed int
                    char str[32];
                    string_dec_to_alpha(str, va_arg(args, int64_t));
                    uint64_t strlen = string_length(str);

                    memory_copy((void*)str, buffer_out, strlen);
                    buffer_out += strlen;
                    break;
                }
                case 'x': {
                    // Hexadecimal int
                    char str[32];

                    buffer_in++;
                    switch(*buffer_in) {
                        case 'b': {
                            // byte
                            string_hex_8bit_to_alpha(str, va_arg(args, int));
                            break;
                        }
                        case 'w': {
                            // word
                            string_hex_16bit_to_alpha(str, va_arg(args, int));
                            break;
                        }
                        case 'd': {
                            // dword
                            string_hex_32bit_to_alpha(str, va_arg(args, uint32_t));
                            break;
                        }
                        case 'q': {
                            // qword
                            string_hex_64bit_to_alpha(str, va_arg(args, uint64_t));
                            break;
                        }
                        default: {
                            string_hex_64bit_to_alpha(str, va_arg(args, uint64_t));
                            buffer_in--;
                            break;
                        }
                    }
                    uint64_t strlen = string_length(str);

                    memory_copy((void*)str, buffer_out, strlen);
                    buffer_out += strlen;
                    break;
                }
                case '?': {
                    char str[7];
                    string_bool_to_alpha(str, (bool)va_arg(args, int));
                    uint64_t strlen = string_length(str);

                    memory_copy((void*)str, buffer_out, strlen);
                    buffer_out += strlen;
                    break;
                }
                case 'b': {
                    if (numeric_specifier == 0) {
                        numeric_specifier = 64;
                    }

                    char str[65];
                    string_bin_to_alpha(str, numeric_specifier, va_arg(args, uint64_t));
                    uint64_t strlen = string_length(str);

                    memory_copy((void*)str, buffer_out, strlen);
                    buffer_out += strlen;
                    break;
                }
            }
        } else {
            *buffer_out = *buffer_in;
            buffer_out++;
        }
        buffer_in++;
    }

    *buffer_out = '\0';
}

void format(string_t output, string_t string, ...) {
    va_list args;
    va_start(args, string);

    variadic_format(output, string, args);

    va_end(args);
}

void string_unsigned_dec_to_alpha(string_t string, uint64_t value) {
    char*       buffer  = (char*)string;
    uint8_t     length  = 0;
    uint64_t    divisor = 10;

    while (value / divisor > 0) {
        length  += 1;
        divisor *= 10;
    }

    divisor = 10;
    for (int i = length; i > 0; i--) {
        buffer[i-1]  = '0' + (value / divisor % 10);
        divisor     *= 10;
    }

    buffer[length]     = '0' + (value % 10);
    buffer[length + 1] = '\0';
}

void string_dec_to_alpha(string_t string, int64_t value) {
    if (value < 0) {
        char* buffer = (char*)string;
        buffer[0] = '-';
        string_unsigned_dec_to_alpha(&string[1], abs(value));
    } else {
        string_unsigned_dec_to_alpha(&string[0], value);
    }
}

void string_hex_8bit_to_alpha(string_t string, uint8_t value) {
    char* buffer = (char*)string;

    buffer[1]   = g_hex_chars[value & 0x0F];
    value     >>= 4;
    buffer[0]   = g_hex_chars[value & 0x0F];
    buffer[2]   = '\0';
}

void string_hex_16bit_to_alpha(string_t string, uint16_t value) {
    char* buffer = (char*)string;

    for (int i = 4; i > 0; i--) {
        buffer[i-1]   = g_hex_chars[value & 0x000F];
        value       >>= 4;
    }
    buffer[4] = '\0';
}

void string_hex_32bit_to_alpha(string_t string, uint32_t value) {
    char* buffer = (char*)string;

    for (int i = 8; i > 0; i--) {
        buffer[i-1]   = g_hex_chars[value & 0x0000000F];
        value       >>= 4;
    }
    buffer[8] = '\0';
}

void string_hex_64bit_to_alpha(string_t string, uint64_t value) {
    char* buffer = (char*)string;

    for (int i = 16; i > 0; i--) {
        buffer[i-1]   = g_hex_chars[value & 0x000000000000000F];
        value       >>= 4;
    }
    buffer[16] = '\0';
}

void string_bin_to_alpha(string_t string, uint8_t num_bits, uint64_t value) {
    char* buffer = (char*)string;

    num_bits   = MIN(num_bits, 64);
    value    <<= 64 - num_bits;
    for (uint8_t i = 0; i < num_bits; i++) {
        buffer[i] = ((value & (0x8000000000000000 >> i)) > 0) ? '1' : '0';
    }
    buffer[num_bits] = '\0';
}

void string_bool_to_alpha(string_t string, bool value) {
    memory_copy(value ? "true" : "false", (void*)string, value ? 5 : 6);
}

bool string_is_char_text(char chr) {
    if (chr >= '!' && chr <= '~') {
        return true;
    }
    return false;
}

bool string_is_char_number(char chr) {
    if (chr >= '0' && chr <= '9') {
        return true;
    }
    return false;
}

bool string_is_char_special(char chr) {
    if ((chr >= '!' && chr <= '/') ||
        (chr >= ':' && chr <= '@') ||
        (chr >= '[' && chr <= '`') ||
        (chr >= '{' && chr <= '~')) {
        return true;
    }
    return false;
}

bool string_is_char_alpha(char chr) {
    if (string_is_char_uppercase(chr) || string_is_char_lowercase(chr) || string_is_char_number(chr)) {
        return true;
    }
    return false;
}

bool string_is_char_uppercase (char chr) {
    if (chr >= 'A' && chr <= 'Z') {
        return true;
    }
    return false;
}

bool string_is_char_lowercase (char chr) {
    if (chr >= 'a' && chr <= 'z') {
        return true;
    }
    return false;
}

uint32_t string_hash_djb(string_t str) {
    uint32_t hash = 5381;

    while (*(++str) != '\0') {
        hash = ((hash << 5) + hash) + (*str);
    }

    return hash;
}
